import asyncio
import datetime
import uuid

import pytest
from pydantic import BaseModel

from langsmith import utils as ls_utils
from langsmith import uuid7
from langsmith.async_client import AsyncClient
from langsmith.schemas import DataType
from langsmith.utils import LangSmithRateLimitError
from tests.integration_tests.conftest import skip_if_rate_limited


@pytest.mark.asyncio
@pytest.mark.skip(reason="Flaky")
async def test_indexed_datasets():
    class InputsSchema(BaseModel):
        name: str  # type: ignore[annotation-unchecked]
        age: int  # type: ignore[annotation-unchecked]

    dataset = None
    async with AsyncClient() as client:
        # Create a new dataset
        try:
            dataset = await client.create_dataset(
                "test_dataset_for_integration_tests_" + uuid.uuid4().hex,
                inputs_schema_definition=InputsSchema.model_json_schema(),
            )

            example = await client.create_example(
                inputs={"name": "Alice", "age": 30},
                outputs={"hi": "hello"},
                dataset_id=dataset.id,
            )

            await client.index_dataset(dataset_id=dataset.id)

            async def check_similar_examples():
                examples = await client.similar_examples(
                    {"name": "Alice", "age": 30}, dataset_id=dataset.id, limit=1
                )
                return len(examples) == 1

            await wait_for(check_similar_examples, timeout=20)
            examples = await client.similar_examples(
                {"name": "Alice", "age": 30}, dataset_id=dataset.id, limit=1
            )
            assert examples[0].id == example.id

            example2 = await client.create_example(
                inputs={"name": "Bobby", "age": 30},
                outputs={"hi": "there"},
                dataset_id=dataset.id,
            )

            await client.sync_indexed_dataset(dataset_id=dataset.id)

            async def check_similar_examples():
                examples = await client.similar_examples(
                    {"name": "Bobby", "age": 30}, dataset_id=dataset.id, limit=2
                )
                return len(examples) == 2

            await wait_for(check_similar_examples, timeout=20)
            examples = await client.similar_examples(
                {"name": "Bobby", "age": 30}, dataset_id=dataset.id, limit=2
            )
            assert examples[0].id == example2.id
            assert examples[1].id == example.id
        finally:
            if dataset:
                await client.delete_dataset(dataset_id=dataset.id)


# Helper function to wait for a condition
async def wait_for(condition, timeout=10, **kwargs):
    start_time = asyncio.get_event_loop().time()
    while True:
        try:
            if await condition(**kwargs):
                return
        except Exception:
            if asyncio.get_event_loop().time() - start_time > timeout:
                raise TimeoutError("Condition not met within the timeout period")
            await asyncio.sleep(0.1)


@pytest.fixture
async def async_client():
    ls_utils.get_env_var.cache_clear()
    client = AsyncClient()
    yield client
    await client.aclose()


@pytest.mark.asyncio
@skip_if_rate_limited
async def test_create_run(async_client: AsyncClient):
    project_name = "__test_create_run" + uuid.uuid4().hex[:8]
    run_id = uuid7()

    await async_client.create_run(
        name="test_run",
        inputs={"input": "hello"},
        run_type="llm",
        project_name=project_name,
        id=run_id,
        start_time=datetime.datetime.now(datetime.timezone.utc),
    )

    async def check_run():
        try:
            run = await async_client.read_run(run_id)
            return run.name == "test_run"
        except ls_utils.LangSmithError:
            return False

    await wait_for(check_run)
    run = await async_client.read_run(run_id)
    assert run.name == "test_run"
    assert run.inputs == {"input": "hello"}


@pytest.mark.asyncio
@skip_if_rate_limited
async def test_list_runs(async_client: AsyncClient):
    project_name = "__test_list_runs" + uuid.uuid4().hex
    run_ids = [uuid7() for _ in range(2)]
    meta_uid = str(uuid7())

    for i, run_id in enumerate(run_ids):
        await async_client.create_run(
            name=f"test_run_{i}",
            inputs={"input": f"hello_{i}"},
            run_type="llm",
            project_name=project_name,
            id=run_id,
            start_time=datetime.datetime.now(datetime.timezone.utc),
            end_time=datetime.datetime.now(datetime.timezone.utc),
            extra={"metadata": {"uid": meta_uid}},
        )

    filter_ = f'and(eq(metadata_key, "uid"), eq(metadata_value, "{meta_uid}"))'

    async def check_runs():
        runs = [
            run
            async for run in async_client.list_runs(
                project_name=project_name, filter=filter_
            )
        ]
        return len(runs) == 2

    try:
        await wait_for(check_runs, timeout=30)
    except TimeoutError:
        raise LangSmithRateLimitError("rate limited")


@pytest.mark.asyncio
async def test_create_dataset(async_client: AsyncClient):
    dataset_name = "__test_create_dataset" + uuid.uuid4().hex[:8]

    dataset = await async_client.create_dataset(dataset_name, data_type=DataType.kv)

    assert dataset.name == dataset_name
    assert dataset.data_type == DataType.kv

    await async_client.delete_dataset(dataset_id=dataset.id)


@pytest.mark.asyncio
async def test_create_example(async_client: AsyncClient):
    dataset_name = "__test_create_example" + uuid.uuid4().hex
    dataset = await async_client.create_dataset(dataset_name)

    example = await async_client.create_example(
        inputs={"input": "hello"}, outputs={"output": "world"}, dataset_id=dataset.id
    )

    assert example.inputs == {"input": "hello"}
    assert example.outputs == {"output": "world"}

    await async_client.delete_dataset(dataset_id=dataset.id)


@pytest.mark.asyncio
async def test_list_examples(async_client: AsyncClient):
    dataset_name = "__test_list_examples" + uuid.uuid4().hex[:8]
    dataset = await async_client.create_dataset(dataset_name)

    for i in range(3):
        await async_client.create_example(
            inputs={"input": f"hello_{i}"},
            outputs={"output": f"world_{i}"},
            dataset_id=dataset.id,
        )

    examples = [
        example async for example in async_client.list_examples(dataset_id=dataset.id)
    ]
    assert len(examples) == 3

    await async_client.delete_dataset(dataset_id=dataset.id)


@pytest.mark.asyncio
@pytest.mark.slow
async def test_create_feedback(async_client: AsyncClient):
    project_name = "__test_create_feedback" + uuid.uuid4().hex[:8]
    run_id = uuid7()

    await async_client.create_run(
        name="test_run",
        inputs={"input": "hello"},
        run_type="llm",
        project_name=project_name,
        id=run_id,
        start_time=datetime.datetime.now(datetime.timezone.utc),
    )

    # Wait for the project to be fully available before creating feedback
    async def check_project_exists():
        try:
            await async_client.read_project(project_name=project_name)
            return True
        except ls_utils.LangSmithError:
            return False

    await wait_for(check_project_exists, timeout=20)

    feedback = await async_client.create_feedback(
        run_id=run_id,
        key="test_key",
        score=0.9,
        value="test_value",
        comment="test_comment",
    )

    assert feedback.run_id == run_id
    assert feedback.key == "test_key"
    assert feedback.score == 0.9
    assert feedback.value == "test_value"
    assert feedback.comment == "test_comment"

    token = await async_client.create_presigned_feedback_token(
        run_id=run_id, feedback_key="test_presigned_key"
    )
    await async_client.create_feedback_from_token(
        token.id, score=0.8, value="presigned_value", comment="presigned_comment"
    )
    await async_client.create_feedback_from_token(
        str(token.url),
        score=0.9,
        value="presigned_value",
        comment="presigned_comment",
    )

    async def check_feedback():
        feedbacks = [
            feedback async for feedback in async_client.list_feedback(run_ids=[run_id])
        ]
        return sum(feedback.key == "test_presigned_key" for feedback in feedbacks) == 2

    await wait_for(check_feedback, timeout=10)
    feedbacks = [
        feedback async for feedback in async_client.list_feedback(run_ids=[run_id])
    ]
    presigned_feedbacks = [f for f in feedbacks if f.key == "test_presigned_key"]
    assert len(presigned_feedbacks) == 2
    assert all(f.value == "presigned_value" for f in presigned_feedbacks)
    assert len(presigned_feedbacks) == 2
    for feedback in presigned_feedbacks:
        assert feedback.value == "presigned_value"
        assert feedback.comment == "presigned_comment"
        assert feedback.score in {0.8, 0.9}
    assert set(f.score for f in presigned_feedbacks) == {0.8, 0.9}

    shared_run_url = None

    # The share endpoint can lag run creation; retry a bit before failing.
    async def check_share():
        nonlocal shared_run_url
        try:
            shared_run_url = await async_client.share_run(run_id)
            return True
        except ls_utils.LangSmithNotFoundError:
            return False

    await wait_for(check_share, timeout=20)
    assert shared_run_url is not None

    run_is_shared = False

    async def check_run_is_shared():
        nonlocal run_is_shared
        try:
            run_is_shared = await async_client.run_is_shared(run_id)
            return run_is_shared
        except ls_utils.LangSmithNotFoundError:
            return False

    await wait_for(check_run_is_shared, timeout=20)
    assert run_is_shared, f"Run isn't shared; failed link: {shared_run_url}"


@pytest.mark.asyncio
async def test_list_feedback(async_client: AsyncClient):
    project_name = "__test_list_feedback"
    run_id = uuid7()

    await async_client.create_run(
        name="test_run",
        inputs={"input": "hello"},
        run_type="llm",
        project_name=project_name,
        id=run_id,
        start_time=datetime.datetime.now(datetime.timezone.utc),
    )

    for i in range(3):
        await async_client.create_feedback(
            run_id=run_id,
            key=f"test_key_{i}",
            score=0.9,
            value=f"test_value_{i}",
            comment=f"test_comment_{i}",
        )

    async def check_feedbacks():
        feedbacks = [
            feedback async for feedback in async_client.list_feedback(run_ids=[run_id])
        ]
        return len(feedbacks) == 3

    await wait_for(check_feedbacks, timeout=10)

    feedbacks = [
        feedback async for feedback in async_client.list_feedback(run_ids=[run_id])
    ]
    assert len(feedbacks) == 3


# TODO: remove skip
@pytest.mark.skip(reason="Flakey")
@pytest.mark.asyncio
async def test_delete_feedback(async_client: AsyncClient):
    """Test deleting feedback."""
    project_name = "__test_delete_feedback" + uuid.uuid4().hex[:8]
    run_id = uuid7()

    await async_client.create_run(
        name="test_run",
        inputs={"input": "hello"},
        run_type="llm",
        project_name=project_name,
        id=run_id,
        start_time=datetime.datetime.now(datetime.timezone.utc),
    )

    # Create feedback
    feedback = await async_client.create_feedback(
        run_id=run_id,
        key="test_feedback",
        value=1,
        comment="test comment",
    )

    # Delete the feedback
    await async_client.delete_feedback(feedback.id)

    # Verify feedback is deleted by checking list_feedback
    feedbacks = [
        feedback async for feedback in async_client.list_feedback(run_ids=[run_id])
    ]
    assert len(feedbacks) == 0


@pytest.mark.asyncio
async def test_annotation_queue_crud(async_client: AsyncClient):
    """Test basic CRUD operations for annotation queues."""
    queue_name = f"test_queue_{uuid.uuid4().hex[:8]}"
    queue_id = uuid7()

    # Test creation
    queue = await async_client.create_annotation_queue(
        name=queue_name, description="Test queue", queue_id=queue_id
    )
    assert queue.name == queue_name
    assert queue.id == queue_id

    # Test reading
    read_queue = await async_client.read_annotation_queue(queue_id)
    assert read_queue.id == queue_id
    assert read_queue.name == queue_name

    # Test updating
    new_name = f"updated_{queue_name}"
    await async_client.update_annotation_queue(
        queue_id=queue_id, name=new_name, description="Updated description"
    )

    updated_queue = await async_client.read_annotation_queue(queue_id)
    assert updated_queue.name == new_name

    # Test deletion
    await async_client.delete_annotation_queue(queue_id)

    # Verify deletion
    queues = [
        queue
        async for queue in async_client.list_annotation_queues(queue_ids=[queue_id])
    ]
    assert len(queues) == 0


@pytest.mark.asyncio
async def test_feedback_config_crud(async_client: AsyncClient):
    """Test basic CRUD operations for feedback configurations."""
    feedback_key = f"test_config_{uuid.uuid4().hex[:8]}"

    # Test creation
    config = await async_client.create_feedback_config(
        feedback_key=feedback_key,
        feedback_config={
            "type": "continuous",
            "min": 0.0,
            "max": 1.0,
        },
        is_lower_score_better=False,
    )
    assert config.feedback_key == feedback_key
    assert config.feedback_config["type"] == "continuous"
    assert config.is_lower_score_better is False

    # Test listing with filter
    configs = [
        c async for c in async_client.list_feedback_configs(feedback_key=[feedback_key])
    ]
    assert len(configs) == 1
    assert configs[0].feedback_key == feedback_key

    # Test update
    updated = await async_client.update_feedback_config(
        feedback_key,
        is_lower_score_better=True,
    )
    assert updated.is_lower_score_better is True

    # Test deletion
    await async_client.delete_feedback_config(feedback_key)

    # Verify deletion
    configs = [
        c async for c in async_client.list_feedback_configs(feedback_key=[feedback_key])
    ]
    assert len(configs) == 0


@pytest.mark.asyncio
async def test_list_annotation_queues(async_client: AsyncClient):
    """Test listing and filtering annotation queues."""
    queue_names = [f"test_queue_{i}_{uuid.uuid4().hex[:8]}" for i in range(3)]
    queue_ids = []

    try:
        # Create test queues
        for name in queue_names:
            queue = await async_client.create_annotation_queue(
                name=name, description="Test queue"
            )
            queue_ids.append(queue.id)

        # Test listing with various filters
        queues = [
            queue
            async for queue in async_client.list_annotation_queues(
                queue_ids=queue_ids[:2], limit=2
            )
        ]
        assert len(queues) == 2

        # Test name filter
        queues = [
            queue
            async for queue in async_client.list_annotation_queues(name=queue_names[0])
        ]
        assert len(queues) == 1
        assert queues[0].name == queue_names[0]

        # Test name_contains filter
        queues = [
            queue
            async for queue in async_client.list_annotation_queues(
                name_contains="test_queue"
            )
        ]
        assert len(queues) >= 3  # Could be more if other tests left queues

    finally:
        # Clean up
        for queue_id in queue_ids:
            await async_client.delete_annotation_queue(queue_id)


@pytest.mark.asyncio
@pytest.mark.slow
async def test_annotation_queue_runs(async_client: AsyncClient):
    """Test managing runs within an annotation queue."""
    queue_name = f"test_queue_{uuid.uuid4().hex[:8]}"
    project_name = f"test_project_{uuid.uuid4().hex[:8]}"

    # Create a queue
    queue = await async_client.create_annotation_queue(
        name=queue_name, description="Test queue"
    )

    # Create some test runs
    run_ids = [uuid7() for _ in range(3)]
    for i, run_id in enumerate(run_ids):
        await async_client.create_run(
            name=f"test_run_{i}",
            inputs={"input": f"test_{i}"},
            run_type="llm",
            project_name=project_name,
            start_time=datetime.datetime.now(datetime.timezone.utc),
            id=run_id,
        )

    async def _get_run(run_id: uuid.UUID) -> bool:
        try:
            await async_client.read_run(run_id)  # type: ignore
            return True
        except ls_utils.LangSmithError:
            return False

    await asyncio.gather(*[wait_for(_get_run, run_id=run_id) for run_id in run_ids])
    # Add runs to queue
    await async_client.add_runs_to_annotation_queue(queue_id=queue.id, run_ids=run_ids)

    # Test getting run at index
    run_info = await async_client.get_run_from_annotation_queue(
        queue_id=queue.id, index=0
    )
    assert run_info.id in run_ids

    # Test deleting a run from queue
    await async_client.delete_run_from_annotation_queue(
        queue_id=queue.id, run_id=run_ids[2]
    )

    # Test that runs are deleted
    with pytest.raises(ls_utils.LangSmithNotFoundError):
        await async_client.get_run_from_annotation_queue(queue_id=queue.id, index=2)

    run_1 = await async_client.get_run_from_annotation_queue(queue_id=queue.id, index=0)
    run_2 = await async_client.get_run_from_annotation_queue(queue_id=queue.id, index=1)
    assert sorted([run_1.id, run_2.id]) == sorted(run_ids[:2])

    # Clean up
    await async_client.delete_annotation_queue(queue.id)



"""Integration tests for Claude Agent SDK tracing."""

import pytest

try:
    import claude_agent_sdk

    CLAUDE_SDK_AVAILABLE = True
except ImportError:
    CLAUDE_SDK_AVAILABLE = False

from langsmith.integrations.claude_agent_sdk._hooks import _active_tool_runs

pytestmark = pytest.mark.skipif(
    not CLAUDE_SDK_AVAILABLE, reason="Claude Agent SDK not installed"
)


@pytest.mark.asyncio
async def test_tool_failure_creates_error_trace():
    """A failing Bash command produces an errored tool run via PostToolUseFailure."""
    from langsmith.integrations.claude_agent_sdk import configure_claude_agent_sdk

    configure_claude_agent_sdk(name="test.tool_failure")

    options = claude_agent_sdk.ClaudeAgentOptions(
        permission_mode="bypassPermissions",
        allowed_tools=["Bash"],
        max_turns=2,
    )

    tool_result_blocks = []
    async with claude_agent_sdk.ClaudeSDKClient(options=options) as client:
        await client.query(
            "Run this exact bash command: cat /tmp/__langsmith_test_nonexistent.txt"
        )
        async for msg in client.receive_response():
            if type(msg).__name__ == "UserMessage" and hasattr(msg, "content"):
                for block in msg.content:
                    if type(block).__name__ == "ToolResultBlock":
                        tool_result_blocks.append(block)

    assert len(tool_result_blocks) >= 1
    assert tool_result_blocks[0].is_error is True

    # PostToolUseFailure hook should have consumed the run — no orphans
    assert len(_active_tool_runs) == 0



from fastapi import FastAPI, Request

from langsmith import traceable
from langsmith.middleware import TracingMiddleware
from langsmith.run_helpers import get_current_run_tree, trace, tracing_context

fake_app = FastAPI()
fake_app.add_middleware(TracingMiddleware)


@traceable
def fake_function():
    span = get_current_run_tree()
    assert span is not None
    assert span.parent_dotted_order
    assert "did-propagate" in span.tags or []
    assert span.metadata["some-cool-value"] == 42
    assert span.session_name == "distributed-tracing"
    return "Fake function response"


@traceable
def fake_function_two(foo: str):
    span = get_current_run_tree()
    assert span is not None
    assert span.parent_dotted_order
    assert "did-propagate" in (span.tags or [])
    assert span.metadata["some-cool-value"] == 42
    assert span.session_name == "distributed-tracing"
    return "Fake function response"


@traceable
def fake_function_three(foo: str):
    span = get_current_run_tree()
    assert span is not None
    assert span.parent_dotted_order
    assert "did-propagate" in (span.tags or [])
    assert span.metadata["some-cool-value"] == 42
    assert span.session_name == "distributed-tracing"
    return "Fake function response"


@fake_app.post("/fake-route")
async def fake_route(request: Request):
    with trace(
        "Trace",
        project_name="Definitely-not-your-grandpas-project",
    ):
        fake_function()
    fake_function_two(
        "foo",
        langsmith_extra={
            "project_name": "Definitely-not-your-grandpas-project",
        },
    )

    with tracing_context(
        parent=request.headers, project_name="Definitely-not-your-grandpas-project"
    ):
        fake_function_three("foo")
    return {"message": "Fake route response"}



import functools
import inspect
import os
import re

import pytest
import vcr

from langsmith import utils as ls_utils


def skip_if_rate_limited(fn):
    """Decorator to skip a test if it raises LangSmithRateLimitError.

    Works for both sync and async tests.
    """

    if inspect.iscoroutinefunction(fn):

        @functools.wraps(fn)
        async def _async_wrapped(*args, **kwargs):
            try:
                return await fn(*args, **kwargs)
            except ls_utils.LangSmithRateLimitError as e:
                pytest.skip(f"LangSmith rate limited: {e}")

        return _async_wrapped

    @functools.wraps(fn)
    def _wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ls_utils.LangSmithRateLimitError as e:
            pytest.skip(f"LangSmith rate limited: {e}")

    return _wrapped


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def is_ai_api_request(request):
    """Check if the request is to OpenAI or Anthropic APIs."""
    ai_domains = [
        "api.openai.com",
        "api.anthropic.com",
    ]

    return any(domain in request.host for domain in ai_domains)


def filter_request_data(request):
    """Filter sensitive data from requests and only record AI API calls."""

    # Only record OpenAI and Anthropic API calls
    if not is_ai_api_request(request):
        return None  # Skip recording this request

    if request.body:
        try:
            # Try to decode as string if it's bytes
            if isinstance(request.body, bytes):
                body_str = request.body.decode("utf-8")
            else:
                body_str = request.body

            # Replace API keys and other sensitive data
            body_str = re.sub(
                r'(api[-_]?key"?\s*:\s*")\w+(")', r"\1FILTERED\2", body_str
            )
            body_str = re.sub(r"(Bearer\s+)[A-Za-z0-9-_]+", r"\1FILTERED", body_str)

            # Update the request body
            if isinstance(request.body, bytes):
                request.body = body_str.encode("utf-8")
            else:
                request.body = body_str
        except (UnicodeDecodeError, AttributeError):
            pass
    return request


def filter_response_data(response):
    """Filter sensitive data from responses."""
    # Only process responses from AI APIs
    if not hasattr(response, "request"):
        return response

    if not is_ai_api_request(response.request):
        return None

    return response


# Create a custom VCR instance with our configuration
def create_vcr_instance(record_mode):
    """Create a configured VCR instance."""
    cassette_dir = os.path.join(os.path.dirname(__file__), "cassettes")

    # Configure VCR
    my_vcr = vcr.VCR(
        cassette_library_dir=cassette_dir,
        record_mode=record_mode,
        match_on=["method", "scheme", "host", "port", "path", "query", "body"],
        filter_headers=[
            "Authorization",
            "X-Api-Key",
            "OpenAI-Organization",
            "Anthropic-Version",
            "User-Agent",
        ],
        # Filter out API keys from request bodies
        before_record_request=filter_request_data,
        before_record_response=filter_response_data,
        decode_compressed_response=True,
    )
    return my_vcr


@pytest.fixture(scope="function", autouse=True)
def vcr_fixture(request):
    """Global VCR fixture that's automatically used for all tests.

    This will record/replay only OpenAI and Anthropic API calls.
    """

    # Get the record mode from command line
    record_mode = request.config.getoption("--vcr-mode")

    # Create the VCR instance
    my_vcr = create_vcr_instance(record_mode)

    # Create a cassette name based on the test function
    module_name = request.module.__name__.split(".")[-1]
    test_name = request.function.__name__
    cassette_name = f"{module_name}_{test_name}"

    # Use the cassette for this test
    with my_vcr.use_cassette(f"{cassette_name}.yaml"):
        yield



import uuid
from unittest.mock import MagicMock

from langsmith.testing._internal import (
    _get_example_id,
    _serde_example_values,
    _TestCase,
)


def test__serde_example_values():
    class Foo:
        def __init__(self, a, b):
            self.a = a
            self.b = b

    results = _serde_example_values({"foo": Foo(1, 2)})
    assert "foo" in results
    assert isinstance(results["foo"], str)


def test__get_id():
    class Foo:
        bar: str = "baz"  # type: ignore

        def __init__(self, a: int, b: int):
            self.a = a
            self.b = b

    def foo(x: Foo):
        return x

    dataset_id = "4e32bff6-5762-4906-8d74-ee2bd0f1d234"
    inputs = {"x": Foo(1, 2)}
    outputs = {"y": Foo(3, 4)}

    result = _get_example_id(dataset_id, inputs, outputs)
    assert isinstance(result, uuid.UUID)


def test_log_inputs_updates_test_case_inputs():
    """Test that calling log_inputs updates _TestCase.inputs.

    When a test has fixture params (e.g. model: BaseChatModel), the initial
    inputs captured from the function signature include the fixture values.
    Calling log_inputs should update self.inputs so that example ID generation
    in end_run uses the correct inputs.
    """
    mock_test_suite = MagicMock()
    mock_test_suite.id = uuid.uuid4()

    test_case = _TestCase(
        test_suite=mock_test_suite,
        run_id=uuid.uuid4(),
        inputs={"model": "<fixture object>", "db": "<db fixture>"},
    )

    # Before log_inputs, inputs reflect fixtures
    assert test_case.inputs == {"model": "<fixture object>", "db": "<db fixture>"}

    # User calls t.log_inputs({"question": "fix typo"}) inside their test.
    # The global log_inputs merges into run_tree then passes run_tree.inputs
    # to test_case.log_inputs. Simulate that merged result:
    test_case.log_inputs({"question": "fix typo"})

    assert test_case.inputs == {"question": "fix typo"}

    # Verify the example ID is now derived from the logged inputs,
    # not the original fixture-based inputs
    dataset_id = str(mock_test_suite.id)
    expected_id = _get_example_id(dataset_id, {"question": "fix typo"})
    fixture_id = _get_example_id(
        dataset_id, {"model": "<fixture object>", "db": "<db fixture>"}
    )
    # end_run falls back to _get_example_id when no explicit id is set
    actual_id = _get_example_id(dataset_id, test_case.inputs or {})
    assert actual_id == expected_id
    assert actual_id != fixture_id



"""Unit tests for Anthropic wrapper processing functions."""

from unittest.mock import MagicMock

from langsmith._internal._orjson import loads as _loads
from langsmith.wrappers._anthropic import _create_usage_metadata, _message_to_outputs


class TestCreateUsageMetadata:
    """Test _create_usage_metadata function."""

    def test_no_cache(self):
        result = _create_usage_metadata({"input_tokens": 100, "output_tokens": 50})
        assert result["input_tokens"] == 100
        assert result["output_tokens"] == 50
        assert result["total_tokens"] == 150
        assert "input_token_details" not in result

    def test_legacy_cache_tokens(self):
        # Anthropic reports cache tokens additively (not as subsets of input_tokens).
        # input_tokens=19 is the NON-cached portion; cache tokens are on top.
        result = _create_usage_metadata(
            {
                "input_tokens": 19,
                "output_tokens": 100,
                "cache_read_input_tokens": 32000,
                "cache_creation_input_tokens": 7000,
            }
        )
        assert result["input_tokens"] == 39019  # 19 + 32000 + 7000
        assert result["output_tokens"] == 100
        assert result["total_tokens"] == 39119  # 39019 + 100
        assert result["input_token_details"]["cache_read"] == 32000
        assert result["input_token_details"]["cache_creation"] == 7000

    def test_new_format_cache_object(self):
        result = _create_usage_metadata(
            {
                "input_tokens": 100,
                "output_tokens": 50,
                "cache_read_input_tokens": 500,
                "cache_creation": {
                    "ephemeral_5m_input_tokens": 800,
                    "ephemeral_1h_input_tokens": 200,
                },
            }
        )
        assert result["input_tokens"] == 1600  # 100 + 500 + 800 + 200
        assert result["output_tokens"] == 50
        assert result["total_tokens"] == 1650  # 1600 + 50
        assert result["input_token_details"]["cache_read"] == 500
        assert result["input_token_details"]["ephemeral_5m_input_tokens"] == 800
        assert result["input_token_details"]["ephemeral_1h_input_tokens"] == 200

    def test_only_cache_read(self):
        result = _create_usage_metadata(
            {
                "input_tokens": 50,
                "output_tokens": 30,
                "cache_read_input_tokens": 1000,
            }
        )
        assert result["input_tokens"] == 1050  # 50 + 1000
        assert result["output_tokens"] == 30
        assert result["total_tokens"] == 1080  # 1050 + 30
        assert result["input_token_details"]["cache_read"] == 1000

    def test_zero_cache_tokens_not_in_details(self):
        # Zero-valued cache fields should not appear in input_token_details
        result = _create_usage_metadata(
            {
                "input_tokens": 100,
                "output_tokens": 50,
                "cache_read_input_tokens": 0,
                "cache_creation_input_tokens": 0,
            }
        )
        assert result["input_tokens"] == 100
        assert result["total_tokens"] == 150
        assert "input_token_details" not in result

    def test_empty_usage(self):
        result = _create_usage_metadata({})
        assert result["input_tokens"] == 0
        assert result["output_tokens"] == 0
        assert result["total_tokens"] == 0


class TestMessageToOutputsToolCalls:
    """Test that _message_to_outputs converts Anthropic tool_use blocks
    to the OpenAI-compatible tool_calls format the LangSmith UI expects."""

    @staticmethod
    def _make_message(content, usage=None, stop_reason="end_turn"):
        msg = MagicMock()
        msg.model_dump.return_value = {
            "id": "msg_123",
            "role": "assistant",
            "content": content,
            "model": "claude-sonnet-4-20250514",
            "stop_reason": stop_reason,
            "type": "message",
            "usage": usage or {"input_tokens": 10, "output_tokens": 5},
        }
        return msg

    def test_tool_use_converted_to_tool_calls(self):
        msg = self._make_message(
            [
                {
                    "type": "tool_use",
                    "id": "toolu_abc",
                    "name": "get_weather",
                    "input": {"location": "SF", "unit": "celsius"},
                },
            ],
            stop_reason="tool_use",
        )
        result = _message_to_outputs(msg)

        assert result["content"] is None
        assert len(result["tool_calls"]) == 1
        tc = result["tool_calls"][0]
        assert tc["id"] == "toolu_abc"
        assert tc["type"] == "function"
        assert tc["index"] == 0
        assert tc["function"]["name"] == "get_weather"
        assert _loads(tc["function"]["arguments"]) == {
            "location": "SF",
            "unit": "celsius",
        }

    def test_text_only_has_no_tool_calls(self):
        msg = self._make_message([{"type": "text", "text": "Hello!"}])
        result = _message_to_outputs(msg)

        assert "tool_calls" not in result
        assert result["content"] == [{"type": "text", "text": "Hello!"}]

    def test_text_stripped_from_content_when_tool_calls_present(self):
        msg = self._make_message(
            [
                {"type": "text", "text": "Let me check."},
                {
                    "type": "tool_use",
                    "id": "toolu_1",
                    "name": "search",
                    "input": {"q": "weather"},
                },
            ],
            stop_reason="tool_use",
        )
        result = _message_to_outputs(msg)

        assert result["content"] == "Let me check."
        assert len(result["tool_calls"]) == 1

    def test_multiple_tool_calls_indexed(self):
        msg = self._make_message(
            [
                {
                    "type": "tool_use",
                    "id": "toolu_1",
                    "name": "get_weather",
                    "input": {"location": "SF"},
                },
                {
                    "type": "tool_use",
                    "id": "toolu_2",
                    "name": "get_stock",
                    "input": {"ticker": "AAPL"},
                },
            ],
            stop_reason="tool_use",
        )
        result = _message_to_outputs(msg)

        assert len(result["tool_calls"]) == 2
        assert result["tool_calls"][0]["index"] == 0
        assert result["tool_calls"][1]["index"] == 1
        assert result["tool_calls"][0]["function"]["name"] == "get_weather"
        assert result["tool_calls"][1]["function"]["name"] == "get_stock"



# mypy: disable-error-code="attr-defined, union-attr, arg-type, call-overload"
import json
import os
import time
from pathlib import Path
from typing import Any
from unittest import mock

import pytest
from pydantic import BaseModel

import langsmith
from langsmith.wrappers import wrap_openai


@pytest.mark.parametrize("stream", [False, True])
def test_chat_sync_api(stream: bool):
    import openai  # noqa

    mock_session = mock.MagicMock()
    client = langsmith.Client(session=mock_session)
    original_client = openai.Client()
    patched_client = wrap_openai(openai.Client(), tracing_extra={"client": client})
    messages = [{"role": "user", "content": "Say 'foo'"}]
    original = original_client.chat.completions.create(
        messages=messages,
        stream=stream,
        temperature=0,
        seed=42,
        model="gpt-3.5-turbo",
    )
    patched = patched_client.chat.completions.create(
        messages=messages,
        stream=stream,
        temperature=0,
        seed=42,
        model="gpt-3.5-turbo",
    )
    if stream:
        # We currently return a generator, so
        # the types aren't the same.
        original_chunks = list(original)
        patched_chunks = list(patched)
        assert len(original_chunks) == len(patched_chunks)
        assert [o.choices == p.choices for o, p in zip(original_chunks, patched_chunks)]
    else:
        assert type(original) is type(patched)
        assert original.choices == patched.choices
    # Give the thread a chance.
    time.sleep(0.01)
    for call in mock_session.request.call_args_list[1:]:
        assert call[0][0].upper() == "POST"


@pytest.mark.parametrize("stream", [False, True])
async def test_chat_async_api(stream: bool):
    import openai  # noqa

    mock_session = mock.MagicMock()
    client = langsmith.Client(session=mock_session)
    original_client = openai.AsyncClient()
    patched_client = wrap_openai(openai.AsyncClient(), tracing_extra={"client": client})
    messages = [{"role": "user", "content": "Say 'foo'"}]
    original = await original_client.chat.completions.create(
        messages=messages, stream=stream, temperature=0, seed=42, model="gpt-3.5-turbo"
    )
    patched = await patched_client.chat.completions.create(
        messages=messages, stream=stream, temperature=0, seed=42, model="gpt-3.5-turbo"
    )
    if stream:
        # We currently return a generator, so
        # the types aren't the same.
        original_chunks = []
        async for chunk in original:
            original_chunks.append(chunk)
        patched_chunks = []
        async for chunk in patched:
            patched_chunks.append(chunk)
        assert len(original_chunks) == len(patched_chunks)
        assert [o.choices == p.choices for o, p in zip(original_chunks, patched_chunks)]
    else:
        assert type(original) is type(patched)
        assert original.choices == patched.choices
    # Give the thread a chance.
    time.sleep(0.1)
    for call in mock_session.request.call_args_list[1:]:
        assert call[0][0].upper() == "POST"


@pytest.mark.parametrize("stream", [False, True])
def test_completions_sync_api(stream: bool):
    import openai

    mock_session = mock.MagicMock()
    client = langsmith.Client(session=mock_session)
    original_client = openai.Client()
    patched_client = wrap_openai(openai.Client(), tracing_extra={"client": client})
    prompt = ("Say 'Foo' then stop.",)
    original = original_client.completions.create(
        model="gpt-3.5-turbo-instruct",
        prompt=prompt,
        max_tokens=3,
        temperature=0,
        seed=42,
        stream=stream,
    )
    patched = patched_client.completions.create(
        model="gpt-3.5-turbo-instruct",
        prompt=prompt,
        max_tokens=3,
        temperature=0,
        seed=42,
        stream=stream,
    )
    if stream:
        # We currently return a generator, so
        # the types aren't the same.
        original_chunks = list(original)
        patched_chunks = list(patched)
        assert len(original_chunks) == len(patched_chunks)
        assert [o.choices == p.choices for o, p in zip(original_chunks, patched_chunks)]
        assert original.response
        assert patched.response
    else:
        assert type(original) is type(patched)
        assert original.choices == patched.choices
    # Give the thread a chance.
    time.sleep(0.1)
    for call in mock_session.request.call_args_list[1:]:
        assert call[0][0].upper() == "POST"


@pytest.mark.parametrize("stream", [False, True])
async def test_completions_async_api(stream: bool):
    import openai

    mock_session = mock.MagicMock()
    client = langsmith.Client(session=mock_session)

    original_client = openai.AsyncClient()
    patched_client = wrap_openai(
        openai.AsyncClient(),
        tracing_extra={"client": client},
        chat_name="chattychat",
        completions_name="incompletions",
    )
    prompt = ("Say 'Hi i'm ChatGPT' then stop.",)
    original = await original_client.completions.create(
        model="gpt-3.5-turbo-instruct",
        prompt=prompt,
        max_tokens=5,
        temperature=0,
        seed=42,
        stream=stream,
    )
    patched = await patched_client.completions.create(
        model="gpt-3.5-turbo-instruct",
        prompt=prompt,
        max_tokens=5,
        temperature=0,
        seed=42,
        stream=stream,
    )
    if stream:
        # We currently return a generator, so
        # the types aren't the same.
        original_chunks = []
        async for chunk in original:
            original_chunks.append(chunk)
        patched_chunks = []
        async for chunk in patched:
            patched_chunks.append(chunk)
        assert len(original_chunks) == len(patched_chunks)
        assert [o.choices == p.choices for o, p in zip(original_chunks, patched_chunks)]
        assert original.response
        assert patched.response
    else:
        assert type(original) is type(patched)
        assert original.choices == patched.choices
    # Give the thread a chance.
    for _ in range(10):
        time.sleep(0.1)
        if mock_session.request.call_count >= 1:
            break
    assert mock_session.request.call_count >= 1
    for call in mock_session.request.call_args_list[1:]:
        assert call[0][0].upper() == "POST"


class Collect:
    """
    Collects the runs for inspection.
    """

    def __init__(self):
        self.run = None

    def __call__(self, run):
        self.run = run


def _collect_requests(mock_session: mock.MagicMock, filename: str):
    mock_requests = mock_session.request.call_args_list
    collected_requests = {}
    for _ in range(10):
        time.sleep(0.1)
        for call in mock_requests:
            if json_bytes := call.kwargs.get("data"):
                json_str = json_bytes.decode("utf-8")
                collected_requests.update(json.loads(json_str))
        all_events = [
            *collected_requests.get("post", []),
            *collected_requests.get("patch", []),
        ]
        # if end_time has been set, we can stop collecting as the background
        # thread has finished processing the run
        if any(event.get("end_time") for event in all_events):
            break
        mock_session.request.call_args_list.clear()

    if os.environ.get("WRITE_TOKEN_COUNTING_TEST_DATA") == "1":
        dir_path = Path(__file__).resolve().parent.parent / "test_data"
        file_path = dir_path / f"{filename}.json"
        with open(file_path, "w") as f:
            json.dump(collected_requests, f, indent=2)


LS_TEST_CLIENT_INFO = {
    "batch_ingest_config": {
        "use_multipart_endpoint": False,
        "scale_up_qsize_trigger": 1000,
        "scale_up_nthreads_limit": 16,
        "scale_down_nempty_trigger": 4,
        "size_limit": 100,
        "size_limit_bytes": 20971520,
    },
}


test_cases = [
    {
        "description": "stream",
        "params": {
            "model": "gpt-4o-mini",
            "messages": [{"role": "user", "content": "howdy"}],
            "stream": True,
            "stream_options": {"include_usage": True},
        },
        "expect_usage_metadata": True,
    },
    {
        "description": "stream no usage",
        "params": {
            "model": "gpt-4o-mini",
            "messages": [{"role": "user", "content": "howdy"}],
            "stream": True,
        },
        "expect_usage_metadata": False,
    },
    {
        "description": "",
        "params": {
            "model": "gpt-4o-mini",
            "messages": [{"role": "user", "content": "howdy"}],
        },
        "expect_usage_metadata": True,
    },
    {
        "description": "reasoning",
        "params": {
            "model": "o4-mini",
            "messages": [
                {
                    "role": "user",
                    "content": (
                        "Write a bash script that takes a matrix represented "
                        "as a string with format '[1,2],[3,4],[5,6]' and prints the "
                        "transpose in the same format."
                    ),
                }
            ],
        },
        "expect_usage_metadata": True,
        "check_reasoning_tokens": True,
    },
    # just test flex as priority can randomly downgrade
    {
        "description": "flex service tier",
        "params": {
            "model": "gpt-5-nano",
            "messages": [{"role": "user", "content": "howdy"}],
            "service_tier": "flex",
        },
        "expect_usage_metadata": True,
        "check_service_tier": "flex",
    },
]


@pytest.mark.parametrize("test_case", test_cases)
def test_wrap_openai_chat_tokens(test_case):
    import openai
    from openai.types.chat import ChatCompletion, ChatCompletionChunk

    oai_client = openai.Client()
    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )
    wrapped_oai_client = wrap_openai(oai_client, tracing_extra={"client": ls_client})

    collect = Collect()
    run_id_to_usage_metadata = {}
    with langsmith.tracing_context(enabled=True):
        params: dict[str, Any] = test_case["params"].copy()
        params["langsmith_extra"] = {"on_end": collect}
        try:
            res = wrapped_oai_client.chat.completions.create(**params)
        except openai.APIConnectionError as e:
            pytest.skip("Openai is having some issues: " + str(e))

        if params.get("stream"):
            for chunk in res:
                assert isinstance(chunk, ChatCompletionChunk)
                if test_case.get("expect_usage_metadata") and hasattr(chunk, "usage"):
                    oai_usage = chunk.usage
        else:
            assert isinstance(res, ChatCompletion)
            oai_usage = res.usage

        if test_case["expect_usage_metadata"]:
            usage_metadata = collect.run.outputs["usage_metadata"]
            assert usage_metadata["input_tokens"] == oai_usage.prompt_tokens
            assert usage_metadata["output_tokens"] == oai_usage.completion_tokens
            assert usage_metadata["total_tokens"] == oai_usage.total_tokens
            if test_case.get("check_reasoning_tokens"):
                assert (
                    usage_metadata["output_token_details"]["reasoning"]
                    == oai_usage.completion_tokens_details.reasoning_tokens
                )
            if test_case.get("check_service_tier"):
                service_tier = test_case["check_service_tier"]
                assert service_tier in usage_metadata["input_token_details"]
                assert service_tier in usage_metadata["output_token_details"]
                assert usage_metadata["input_token_details"][service_tier] > 0
                assert usage_metadata["output_token_details"][service_tier] > 0
        else:
            assert collect.run.outputs.get("usage_metadata") is None
            assert collect.run.outputs.get("usage") is None

        run_id_to_usage_metadata[collect.run.id] = collect.run.outputs.get(
            "usage_metadata"
        )

    filename = f"langsmith_py_wrap_openai_{test_case['description'].replace(' ', '_')}"
    _collect_requests(mock_session, filename)


@pytest.mark.asyncio
@pytest.mark.parametrize("test_case", test_cases)
async def test_wrap_openai_chat_async_tokens(test_case):
    import openai
    from openai.types.chat import ChatCompletion, ChatCompletionChunk

    oai_client = openai.AsyncClient()
    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )
    wrapped_oai_client = wrap_openai(oai_client, tracing_extra={"client": ls_client})

    collect = Collect()
    run_id_to_usage_metadata = {}
    with langsmith.tracing_context(enabled=True):
        params: dict[str, Any] = test_case["params"].copy()
        params["langsmith_extra"] = {"on_end": collect}
        try:
            res = await wrapped_oai_client.chat.completions.create(**params)
        except openai.APIConnectionError:
            pytest.skip("Openai is having some issues.")

        if params.get("stream"):
            oai_usage = None
            async for chunk in res:
                assert isinstance(chunk, ChatCompletionChunk)
                if test_case.get("expect_usage_metadata") and hasattr(chunk, "usage"):
                    oai_usage = chunk.usage
        else:
            assert isinstance(res, ChatCompletion)
            oai_usage = res.usage

        if test_case["expect_usage_metadata"]:
            usage_metadata = collect.run.outputs["usage_metadata"]
            assert usage_metadata["input_tokens"] == oai_usage.prompt_tokens
            assert usage_metadata["output_tokens"] == oai_usage.completion_tokens
            assert usage_metadata["total_tokens"] == oai_usage.total_tokens
            if test_case.get("check_reasoning_tokens"):
                assert (
                    usage_metadata["output_token_details"]["reasoning"]
                    == oai_usage.completion_tokens_details.reasoning_tokens
                )
            if test_case.get("check_service_tier"):
                service_tier = test_case["check_service_tier"]
                assert service_tier in usage_metadata["input_token_details"]
                assert service_tier in usage_metadata["output_token_details"]
                assert usage_metadata["input_token_details"][service_tier] > 0
                assert usage_metadata["output_token_details"][service_tier] > 0
        else:
            assert collect.run.outputs.get("usage_metadata") is None
            assert collect.run.outputs.get("usage") is None

        run_id_to_usage_metadata[collect.run.id] = collect.run.outputs.get(
            "usage_metadata"
        )

    filename = f"langsmith_py_wrap_openai_{test_case['description'].replace(' ', '_')}"
    _collect_requests(mock_session, filename)


def test_parse_sync_api():
    """Test that the sync parse method can be traced without errors."""
    import openai  # noqa

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )

    original_client = openai.Client()
    patched_client = wrap_openai(openai.Client(), tracing_extra={"client": ls_client})

    messages = [{"role": "user", "content": "Say 'foo' then stop."}]

    # beta parse
    original_beta = original_client.beta.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )
    patched_beta = patched_client.beta.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )
    # chat parse
    original = original_client.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )
    patched = patched_client.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )

    assert type(original_beta) is type(patched_beta)
    assert original_beta.choices == patched_beta.choices
    assert type(original) is type(patched)
    assert original.choices == patched.choices

    time.sleep(0.1)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]

    _collect_requests(mock_session, "test_parse_sync_api")


@pytest.mark.asyncio
async def test_parse_async_api():
    """Test that the async parse method can be traced without errors."""
    import openai  # noqa

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )

    original_client = openai.AsyncClient()
    patched_client = wrap_openai(
        openai.AsyncClient(), tracing_extra={"client": ls_client}
    )

    messages = [{"role": "user", "content": "Say 'foo' then stop."}]

    original_beta = await original_client.beta.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )
    patched_beta = await patched_client.beta.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )
    original = await original_client.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )
    patched = await patched_client.chat.completions.parse(
        messages=messages, model="gpt-3.5-turbo", temperature=0, seed=42, max_tokens=3
    )

    assert type(original_beta) is type(patched_beta)
    assert original_beta.choices == patched_beta.choices
    assert original.choices == patched.choices
    assert type(original) is type(patched)

    time.sleep(0.1)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]

    _collect_requests(mock_session, "test_parse_async_api")


def test_parse_tokens():
    """
    Test that usage tokens are captured for parse calls
    """
    import openai
    from openai.types.chat import ChatCompletion

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )
    wrapped_oai_client = wrap_openai(
        openai.Client(), tracing_extra={"client": ls_client}
    )

    collect = Collect()
    messages = [{"role": "user", "content": "Check usage"}]

    with langsmith.tracing_context(enabled=True):
        res = wrapped_oai_client.beta.chat.completions.parse(
            messages=messages,
            model="gpt-3.5-turbo",
            langsmith_extra={"on_end": collect},
        )
        assert isinstance(res, ChatCompletion)

    usage_metadata = collect.run.outputs.get("usage_metadata")

    if usage_metadata:
        assert usage_metadata["input_tokens"] >= 0
        assert usage_metadata["output_tokens"] >= 0
        assert usage_metadata["total_tokens"] >= 0

    time.sleep(0.1)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]

    _collect_requests(mock_session, "test_parse_tokens")


def test_responses_sync_api():
    """Test that the sync responses methods can be traced without errors."""
    import openai  # noqa

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )

    original_client = openai.Client()
    patched_client = wrap_openai(openai.Client(), tracing_extra={"client": ls_client})

    # Test create
    original_create = original_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=16,
    )
    patched_create = patched_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=16,
    )
    assert type(original_create) is type(patched_create)
    assert original_create.output_text.lower().strip(
        " ."
    ) == patched_create.output_text.lower().strip(" .")

    # Test create(stream=True)
    original_create_stream = original_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=16,
        stream=True,
    )
    patched_create_stream = patched_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=16,
        stream=True,
    )
    originals, patched = [], []
    for original_event, patched_event in zip(
        original_create_stream, patched_create_stream
    ):
        originals.append(original_event)
        patched.append(patched_event)
    assert len(originals) == len(patched)

    class Format(BaseModel):
        speech: str

    # Test parse
    original_parse = original_client.responses.parse(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=32,
        text_format=Format,
    )
    patched_parse = patched_client.responses.parse(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=32,
        text_format=Format,
    )

    assert isinstance(original_parse.output[0].content[0].parsed, Format)
    assert isinstance(patched_parse.output[0].content[0].parsed, Format)
    assert (
        original_parse.output[0].content[0].parsed
        == patched_parse.output[0].content[0].parsed
    )

    time.sleep(0.1)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]

    _collect_requests(mock_session, "test_responses_sync_api")


@pytest.mark.asyncio
async def test_responses_async_api():
    """Test that the async responses methods can be traced without errors."""
    import openai  # noqa

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )

    original_client = openai.AsyncClient()
    patched_client = wrap_openai(
        openai.AsyncClient(), tracing_extra={"client": ls_client}
    )

    # Test create
    original_create = await original_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=16,
    )
    patched_create = await patched_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=16,
    )
    assert type(original_create) is type(patched_create)
    assert original_create.output_text.lower().strip(
        " ."
    ) == patched_create.output_text.lower().strip(" .")

    # Test stream
    original_stream = await original_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        stream=True,
        temperature=0,
        max_output_tokens=16,
    )
    patched_stream = await patched_client.responses.create(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        stream=True,
        temperature=0,
        max_output_tokens=16,
    )
    original_chunks = []
    async for chunk in original_stream:
        original_chunks.append(chunk)
    patched_chunks = []
    async for chunk in patched_stream:
        patched_chunks.append(chunk)
    assert len(original_chunks) == len(patched_chunks)
    for orig, patched in zip(original_chunks, patched_chunks):
        assert orig.type == patched.type

    # Test parse
    class Format(BaseModel):
        speech: str

    # Test parse
    original_parse = await original_client.responses.parse(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=1024,
        text_format=Format,
    )
    patched_parse = await patched_client.responses.parse(
        input="Say 'foo' then stop.",
        model="gpt-4o-mini",
        temperature=0,
        max_output_tokens=1024,
        text_format=Format,
    )

    assert isinstance(original_parse.output[0].content[0].parsed, Format)
    assert isinstance(patched_parse.output[0].content[0].parsed, Format)
    assert (
        original_parse.output[0].content[0].parsed
        == patched_parse.output[0].content[0].parsed
    )

    time.sleep(0.1)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]

    _collect_requests(mock_session, "test_responses_sync_api")


def test_responses_stream_sync_api():
    """Test that the sync responses methods can be traced without errors."""
    import openai  # noqa

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )
    original_client = openai.Client()
    patched_client = wrap_openai(openai.Client(), tracing_extra={"client": ls_client})

    # Test stream
    with original_client.responses.stream(
        input="Say 'foo' then stop.",
        model="gpt-5-nano",
        max_output_tokens=1024,
    ) as original_stream:
        for _ in original_stream:
            pass
        original_full = original_stream.get_final_response()
    with patched_client.responses.stream(
        input="Say 'foo' then stop.",
        model="gpt-5-nano",
        max_output_tokens=1024,
    ) as patched_stream:
        for _ in patched_stream:
            pass
        patched_full = patched_stream.get_final_response()
    original_chunks = list(original_stream)
    patched_chunks = list(patched_stream)
    assert len(original_chunks) == len(patched_chunks)
    for orig, patched in zip(original_chunks, patched_chunks):
        assert orig.output_text == patched.output_text
    assert original_full.output_text == patched_full.output_text

    time.sleep(0.5)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]

    _collect_requests(mock_session, "test_responses_stream_sync_api")


async def test_responses_stream_async_api():
    """Test that the async responses methods can be traced without errors."""
    import openai  # noqa

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )
    original_client = openai.AsyncClient()
    patched_client = wrap_openai(
        openai.AsyncClient(), tracing_extra={"client": ls_client}
    )
    original_chunks = []
    patched_chunks = []
    async with original_client.responses.stream(
        input="Say 'foo' then stop.",
        model="gpt-5-nano",
        max_output_tokens=1024,
    ) as original_stream:
        async for chunk in original_stream:
            original_chunks.append(chunk)
            pass
        original_full = await original_stream.get_final_response()
    async with patched_client.responses.stream(
        input="Say 'foo' then stop.",
        model="gpt-5-nano",
        max_output_tokens=1024,
    ) as patched_stream:
        async for chunk in patched_stream:
            patched_chunks.append(chunk)
            pass
        patched_full = await patched_stream.get_final_response()
    assert len(original_chunks) == len(patched_chunks)
    assert original_full.output_text == patched_full.output_text

    time.sleep(0.5)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]

    _collect_requests(mock_session, "test_responses_stream_async_api")


@pytest.mark.asyncio
async def test_tool_call_chunking():
    """Test that wrap_openai can reduce tool call chunks when streaming."""
    import openai
    from openai.types.chat import ChatCompletionChunk

    mock_session = mock.MagicMock()
    ls_client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )

    client = wrap_openai(openai.AsyncClient(), tracing_extra={"client": ls_client})

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get the current weather in a given city.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city to get the weather for",
                        }
                    },
                    "required": ["location"],
                },
            },
        }
    ]

    # Simulate a user message
    messages = [{"role": "user", "content": "What's the weather like in Paris?"}]

    collect = Collect()
    with langsmith.tracing_context(enabled=True):
        chunks = await client.chat.completions.create(
            model="gpt-4o",
            messages=messages,
            tools=tools,
            tool_choice="auto",
            stream=True,
            langsmith_extra={"on_end": collect},
        )

        tool_call_chunks = []
        tool_call_ids = set()
        async for chunk in chunks:
            assert isinstance(chunk, ChatCompletionChunk)
            # Collect chunks that contain tool calls
            for choice in chunk.choices:
                if choice.delta.tool_calls:
                    tool_call_chunks.append(chunk)
                    for tool_call in choice.delta.tool_calls:
                        if tool_call.id:
                            tool_call_ids.add(tool_call.id)

        # Verify that we have tool call chunks
        assert len(tool_call_chunks) > 0
        # Verify that there are tool call IDs
        assert len(tool_call_ids) > 0

        # Verify that the run has the expected data
        assert collect.run
        assert collect.run.outputs["choices"][0]["finish_reason"] == "tool_calls"
        assert collect.run.outputs["choices"][0]["message"]["tool_calls"]
        assert (
            collect.run.outputs["choices"][0]["message"]["tool_calls"][0]["function"][
                "name"
            ]
            == "get_weather"
        )

    # Allow time for background threads to complete
    time.sleep(0.1)
    for call in mock_session.request.call_args_list:
        assert call[0][0].upper() in ["POST", "GET", "PATCH"]


def test_prepopulated_invocation_params():
    """Test that prepopulated invocation params are merged
    and runtime params override."""
    import openai

    mock_session = mock.MagicMock()
    client = langsmith.Client(
        session=mock_session,
        info=LS_TEST_CLIENT_INFO,
    )

    # Wrap client with prepopulated params including seed
    patched_client = wrap_openai(
        openai.Client(),
        tracing_extra={
            "client": client,
            "metadata": {
                "ls_invocation_params": {"seed": 100, "test": "qa"},
                "custom_key": "custom_value",
                "version": "1.0.0",
            },
        },
    )

    collect = Collect()
    messages = [{"role": "user", "content": "Say 'hello'"}]

    with langsmith.tracing_context(enabled=True):
        patched_client.chat.completions.create(
            messages=messages,
            seed=42,  # Should override prepopulated seed=100
            model="gpt-5-nano",
            langsmith_extra={"on_end": collect},
        )

    # Give the thread a chance
    time.sleep(0.1)

    # Check the run data from the callback
    assert collect.run is not None

    # Check invocation params - they are in metadata, not extra.invocation_params
    metadata = collect.run.extra.get("metadata", {})
    ls_invocation_params = metadata.get("ls_invocation_params", {})

    # Runtime seed should override prepopulated seed
    assert ls_invocation_params.get("seed") == 42
    # Prepopulated params without conflicts should still be there
    assert ls_invocation_params.get("test") == "qa"

    # Check that other metadata keys are preserved
    assert metadata.get("custom_key") == "custom_value"
    assert metadata.get("version") == "1.0.0"



"""Unit tests for Claude Agent SDK hooks."""

import asyncio
import sys
from unittest.mock import MagicMock

import pytest

from langsmith.integrations.claude_agent_sdk._hooks import (
    _active_tool_runs,
    _client_managed_runs,
    post_tool_use_failure_hook,
    post_tool_use_hook,
    pre_tool_use_hook,
)
from langsmith.run_trees import RunTree

ERROR_MSG = "Exit code 1\ncat: /nonexistent: No such file or directory"


@pytest.fixture(autouse=True)
def _clear_state():
    """Reset global hook state between tests."""
    _active_tool_runs.clear()
    _client_managed_runs.clear()
    yield
    _active_tool_runs.clear()
    _client_managed_runs.clear()


def _make_parent_run() -> RunTree:
    """Create a detached RunTree suitable for parenting child runs."""
    return RunTree(name="test-parent", run_type="chain", client=MagicMock())


class TestToolUseSuccessFlow:
    """PreToolUse creates a child run; PostToolUse ends it with output."""

    @pytest.fixture(autouse=True)
    def _set_parent(self):
        from langsmith.integrations.claude_agent_sdk import _tools

        _tools.set_parent_run_tree(_make_parent_run())
        yield
        _tools.clear_parent_run_tree()

    def test_success_flow(self):
        asyncio.run(
            pre_tool_use_hook(
                {"tool_name": "Bash", "tool_input": {"command": "echo hi"}},
                "tu_1",
                MagicMock(),
            )
        )

        assert "tu_1" in _active_tool_runs
        tool_run, _ = _active_tool_runs["tu_1"]
        assert tool_run.name == "Bash"
        assert tool_run.run_type == "tool"
        assert tool_run.inputs == {"input": {"command": "echo hi"}}

        asyncio.run(
            post_tool_use_hook(
                {
                    "tool_name": "Bash",
                    "tool_response": {"output": "hi", "is_error": False},
                },
                "tu_1",
                MagicMock(),
            )
        )

        assert "tu_1" not in _active_tool_runs
        assert tool_run.outputs == {"output": "hi", "is_error": False}
        assert tool_run.error is None


class TestToolUseFailureFlow:
    """PreToolUse creates a child run; PostToolUseFailure marks it as errored."""

    @pytest.fixture(autouse=True)
    def _set_parent(self):
        from langsmith.integrations.claude_agent_sdk import _tools

        _tools.set_parent_run_tree(_make_parent_run())
        yield
        _tools.clear_parent_run_tree()

    def test_failure_flow(self):
        asyncio.run(
            pre_tool_use_hook(
                {
                    "tool_name": "Bash",
                    "tool_input": {"command": "cat /nonexistent"},
                },
                "tu_2",
                MagicMock(),
            )
        )

        tool_run, _ = _active_tool_runs["tu_2"]

        asyncio.run(
            post_tool_use_failure_hook(
                {
                    "tool_name": "Bash",
                    "tool_input": {"command": "cat /nonexistent"},
                    "error": ERROR_MSG,
                },
                "tu_2",
                MagicMock(),
            )
        )

        assert "tu_2" not in _active_tool_runs
        assert tool_run.error == ERROR_MSG
        assert tool_run.outputs == {"error": ERROR_MSG}


class TestInjectTracingHooks:
    def test_injects_all_three_hooks(self):
        from langsmith.integrations.claude_agent_sdk._client import (
            _inject_tracing_hooks,
        )

        options = MagicMock()
        options.hooks = None

        mock_module = MagicMock()
        sys.modules["claude_agent_sdk"] = mock_module
        try:
            _inject_tracing_hooks(options)
        finally:
            del sys.modules["claude_agent_sdk"]

        for event in ("PreToolUse", "PostToolUse", "PostToolUseFailure"):
            assert event in options.hooks
            assert len(options.hooks[event]) == 1



from typing import Literal
from uuid import uuid4

import pytest
from langchain_core.prompts import (
    BasePromptTemplate,
    ChatPromptTemplate,
    PromptTemplate,
)
from langchain_core.runnables.base import RunnableSequence

import langsmith.schemas as ls_schemas
import langsmith.utils as ls_utils
from langsmith.client import (
    Client,
    convert_prompt_to_anthropic_format,
    convert_prompt_to_openai_format,
)
from tests.integration_tests.conftest import skip_if_rate_limited
from tests.utils.any_int import AnyInt


@pytest.fixture
def langsmith_client() -> Client:
    return Client(timeout_ms=(50_000, 90_000))


@pytest.fixture
def prompt_template_1() -> ChatPromptTemplate:
    return ChatPromptTemplate.from_template("tell me a joke about {topic}")


@pytest.fixture
def prompt_template_2() -> ChatPromptTemplate:
    return ChatPromptTemplate.from_messages(
        [
            ("system", "You are a helpful assistant."),
            ("human", "{question}"),
        ]
    )


@pytest.fixture
def prompt_template_3() -> PromptTemplate:
    return PromptTemplate.from_template("Summarize the following text: {text}")


@pytest.fixture
def prompt_with_model() -> dict:
    return {
        "id": ["langsmith", "playground", "PromptPlayground"],
        "lc": 1,
        "type": "constructor",
        "kwargs": {
            "last": {
                "id": ["langchain", "schema", "runnable", "RunnableBinding"],
                "lc": 1,
                "type": "constructor",
                "kwargs": {
                    "bound": {
                        "id": ["langchain", "chat_models", "openai", "ChatOpenAI"],
                        "lc": 1,
                        "type": "constructor",
                        "kwargs": {
                            "openai_api_key": {
                                "id": ["OPENAI_API_KEY"],
                                "lc": 1,
                                "type": "secret",
                            }
                        },
                    },
                    "kwargs": {},
                },
            },
            "first": {
                "id": ["langchain", "prompts", "chat", "ChatPromptTemplate"],
                "lc": 1,
                "type": "constructor",
                "kwargs": {
                    "messages": [
                        {
                            "id": [
                                "langchain",
                                "prompts",
                                "chat",
                                "SystemMessagePromptTemplate",
                            ],
                            "lc": 1,
                            "type": "constructor",
                            "kwargs": {
                                "prompt": {
                                    "id": [
                                        "langchain",
                                        "prompts",
                                        "prompt",
                                        "PromptTemplate",
                                    ],
                                    "lc": 1,
                                    "type": "constructor",
                                    "kwargs": {
                                        "template": "You are a chatbot.",
                                        "input_variables": [],
                                        "template_format": "f-string",
                                    },
                                }
                            },
                        },
                        {
                            "id": [
                                "langchain",
                                "prompts",
                                "chat",
                                "HumanMessagePromptTemplate",
                            ],
                            "lc": 1,
                            "type": "constructor",
                            "kwargs": {
                                "prompt": {
                                    "id": [
                                        "langchain",
                                        "prompts",
                                        "prompt",
                                        "PromptTemplate",
                                    ],
                                    "lc": 1,
                                    "type": "constructor",
                                    "kwargs": {
                                        "template": "{question}",
                                        "input_variables": ["question"],
                                        "template_format": "f-string",
                                    },
                                }
                            },
                        },
                    ],
                    "input_variables": ["question"],
                },
            },
        },
    }


@pytest.fixture
def chat_prompt_template():
    return ChatPromptTemplate.from_messages(
        [
            ("system", "You are a chatbot"),
            ("user", "{question}"),
        ]
    )


def test_current_tenant_is_owner(langsmith_client: Client):
    settings = langsmith_client._get_settings()
    assert langsmith_client._current_tenant_is_owner(settings.tenant_handle or "-")
    assert langsmith_client._current_tenant_is_owner("-")
    assert not langsmith_client._current_tenant_is_owner("non_existent_owner")


@skip_if_rate_limited
def test_list_prompts(langsmith_client: Client):
    response = langsmith_client.list_prompts(limit=10, offset=0)
    assert isinstance(response, ls_schemas.ListPromptsResponse)
    assert len(response.repos) <= 10


@skip_if_rate_limited
def test_get_prompt(langsmith_client: Client, prompt_template_1: ChatPromptTemplate):
    prompt_name = f"test_prompt_{uuid4().hex[:8]}"
    url = langsmith_client.push_prompt(prompt_name, object=prompt_template_1)
    assert isinstance(url, str)
    assert langsmith_client._prompt_exists(prompt_name)

    prompt = langsmith_client.get_prompt(prompt_name)
    assert isinstance(prompt, ls_schemas.Prompt)
    assert prompt.repo_handle == prompt_name

    langsmith_client.delete_prompt(prompt_name)


@skip_if_rate_limited
def test_prompt_exists(langsmith_client: Client, prompt_template_2: ChatPromptTemplate):
    non_existent_prompt = f"non_existent_{uuid4().hex[:8]}"
    assert not langsmith_client._prompt_exists(non_existent_prompt)

    existent_prompt = f"existent_{uuid4().hex[:8]}"
    assert langsmith_client.push_prompt(existent_prompt, object=prompt_template_2)
    assert langsmith_client._prompt_exists(existent_prompt)

    langsmith_client.delete_prompt(existent_prompt)


@skip_if_rate_limited
def test_update_prompt(langsmith_client: Client, prompt_template_1: ChatPromptTemplate):
    prompt_name = f"test_prompt_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(prompt_name, object=prompt_template_1)

    updated_data = langsmith_client.update_prompt(
        prompt_name,
        description="Updated description",
        is_public=True,
        tags=["test", "update"],
    )
    assert isinstance(updated_data, dict)

    updated_prompt = langsmith_client.get_prompt(prompt_name)
    assert isinstance(updated_prompt, ls_schemas.Prompt)
    assert updated_prompt.description == "Updated description"
    assert updated_prompt.is_public
    assert set(updated_prompt.tags) == set(["test", "update"])

    langsmith_client.delete_prompt(prompt_name)


def test_delete_prompt(langsmith_client: Client, prompt_template_1: ChatPromptTemplate):
    prompt_name = f"test_prompt_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(prompt_name, object=prompt_template_1)

    assert langsmith_client._prompt_exists(prompt_name)
    langsmith_client.delete_prompt(prompt_name)
    assert not langsmith_client._prompt_exists(prompt_name)


def test_pull_prompt_object(
    langsmith_client: Client, prompt_template_1: ChatPromptTemplate
):
    prompt_name = f"test_prompt_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(prompt_name, object=prompt_template_1)

    manifest = langsmith_client.pull_prompt_commit(prompt_name)
    assert isinstance(manifest, ls_schemas.PromptCommit)
    assert manifest.repo == prompt_name

    langsmith_client.delete_prompt(prompt_name)


@skip_if_rate_limited
def test_pull_prompt(langsmith_client: Client, prompt_template_1: ChatPromptTemplate):
    prompt_name = f"test_prompt_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(prompt_name, object=prompt_template_1)

    # test pulling with just prompt name
    pulled_prompt = langsmith_client.pull_prompt(prompt_name)
    assert isinstance(pulled_prompt, ChatPromptTemplate)
    assert (
        pulled_prompt.metadata and pulled_prompt.metadata["lc_hub_repo"] == prompt_name
    )

    # test pulling with private owner (-) and name
    pulled_prompt_2 = langsmith_client.pull_prompt(f"-/{prompt_name}")
    assert pulled_prompt == pulled_prompt_2

    # test pulling with tenant handle and name
    tenant_handle = langsmith_client._get_settings().tenant_handle
    pulled_prompt_3 = langsmith_client.pull_prompt(f"{tenant_handle}/{prompt_name}")
    assert pulled_prompt.metadata and pulled_prompt_3.metadata
    assert (
        pulled_prompt.metadata["lc_hub_commit_hash"]
        == pulled_prompt_3.metadata["lc_hub_commit_hash"]
    )
    assert pulled_prompt_3.metadata["lc_hub_owner"] == tenant_handle

    # test pulling with handle, name and commit hash
    tenant_handle = langsmith_client._get_settings().tenant_handle
    pulled_prompt_4 = langsmith_client.pull_prompt(
        f"{tenant_handle}/{prompt_name}:latest"
    )
    assert pulled_prompt_3 == pulled_prompt_4

    # test pulling without handle, with commit hash
    assert pulled_prompt_4.metadata
    pulled_prompt_5 = langsmith_client.pull_prompt(
        f"{prompt_name}:{pulled_prompt_4.metadata['lc_hub_commit_hash']}"
    )
    assert pulled_prompt_5.metadata
    assert (
        pulled_prompt_4.metadata["lc_hub_commit_hash"]
        == pulled_prompt_5.metadata["lc_hub_commit_hash"]
    )

    langsmith_client.delete_prompt(prompt_name)


@skip_if_rate_limited
def test_push_and_pull_prompt(
    langsmith_client: Client, prompt_template_2: ChatPromptTemplate
):
    prompt_name = f"test_prompt_{uuid4().hex[:8]}"

    push_result = langsmith_client.push_prompt(prompt_name, object=prompt_template_2)
    assert isinstance(push_result, str)

    pulled_prompt = langsmith_client.pull_prompt(prompt_name)
    assert isinstance(pulled_prompt, ChatPromptTemplate)

    langsmith_client.delete_prompt(prompt_name)

    # should fail
    with pytest.raises(ls_utils.LangSmithUserError):
        langsmith_client.push_prompt(
            f"random_handle/{prompt_name}", object=prompt_template_2
        )


@skip_if_rate_limited
def test_pull_prompt_include_model(langsmith_client: Client, prompt_with_model: dict):
    prompt_name = f"test_prompt_with_model_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(prompt_name, object=prompt_with_model)

    pulled_prompt = langsmith_client.pull_prompt(prompt_name, include_model=True)
    assert isinstance(pulled_prompt, RunnableSequence)
    if getattr(pulled_prompt, "first", None):
        first = getattr(pulled_prompt, "first")
        assert isinstance(first, BasePromptTemplate)
        assert first.metadata and first.metadata["lc_hub_repo"] == prompt_name
    else:
        assert False, "pulled_prompt.first should exist, incorrect prompt format"

    langsmith_client.delete_prompt(prompt_name)


@skip_if_rate_limited
def test_get_latest_commit_hash(
    langsmith_client: Client, prompt_template_1: ChatPromptTemplate
):
    prompt_name = f"test_prompt_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(prompt_name, object=prompt_template_1)

    commit_hash = langsmith_client._get_latest_commit_hash(f"-/{prompt_name}")
    assert isinstance(commit_hash, str)
    assert len(commit_hash) > 0

    langsmith_client.delete_prompt(prompt_name)


def test_create_prompt(langsmith_client: Client):
    prompt_name = f"test_create_prompt_{uuid4().hex[:8]}"
    created_prompt = langsmith_client.create_prompt(
        prompt_name,
        description="Test description",
        readme="Test readme",
        tags=["test", "create"],
        is_public=False,
    )
    assert isinstance(created_prompt, ls_schemas.Prompt)
    assert created_prompt.repo_handle == prompt_name
    assert created_prompt.description == "Test description"
    assert created_prompt.readme == "Test readme"
    assert set(created_prompt.tags) == set(["test", "create"])
    assert not created_prompt.is_public

    langsmith_client.delete_prompt(prompt_name)


@skip_if_rate_limited
def test_create_commit(
    langsmith_client: Client,
    prompt_template_2: ChatPromptTemplate,
    prompt_template_3: PromptTemplate,
):
    prompt_name = f"test_create_commit_{uuid4().hex[:8]}"
    try:
        # this should fail because the prompt does not exist
        commit_url = langsmith_client.create_commit(
            prompt_name, object=prompt_template_2
        )
        pytest.fail("Expected LangSmithNotFoundError was not raised")
    except ls_utils.LangSmithNotFoundError as e:
        assert str(e) == "Prompt does not exist, you must create it first."
    except Exception as e:
        pytest.fail(f"Unexpected exception raised: {e}")

    langsmith_client.push_prompt(prompt_name, object=prompt_template_3)
    commit_url = langsmith_client.create_commit(prompt_name, object=prompt_template_2)
    assert isinstance(commit_url, str)
    assert prompt_name in commit_url

    prompt = langsmith_client.get_prompt(prompt_name)
    assert isinstance(prompt, ls_schemas.Prompt)
    assert prompt.num_commits == 2

    # try submitting different types of unaccepted manifests
    try:
        # this should fail
        commit_url = langsmith_client.create_commit(prompt_name, object={"hi": "hello"})
    except ls_utils.LangSmithError as e:
        err = str(e)
        assert "Manifest must have an id field" in err
        assert "400 Client Error" in err
    except Exception as e:
        pytest.fail(f"Unexpected exception raised: {e}")

    try:
        # this should fail
        commit_url = langsmith_client.create_commit(prompt_name, object={"id": ["hi"]})
    except ls_utils.LangSmithError as e:
        err = str(e)
        assert "Manifest type hi is not supported" in err
        assert "400 Client Error" in err
    except Exception as e:
        pytest.fail(f"Unexpected exception raised: {e}")

    langsmith_client.delete_prompt(prompt_name)


@skip_if_rate_limited
def test_push_prompt(
    langsmith_client: Client,
    prompt_template_3: PromptTemplate,
    prompt_template_2: ChatPromptTemplate,
):
    prompt_name = f"test_push_new_{uuid4().hex[:8]}"
    url = langsmith_client.push_prompt(
        prompt_name,
        object=prompt_template_3,
        is_public=True,
        description="New prompt",
        tags=["new", "test"],
    )

    assert isinstance(url, str)
    assert prompt_name in url

    prompt = langsmith_client.get_prompt(prompt_name)
    assert isinstance(prompt, ls_schemas.Prompt)
    assert prompt.is_public
    assert prompt.description == "New prompt"
    assert "new" in prompt.tags
    assert "test" in prompt.tags
    assert prompt.num_commits == 1

    # test updating prompt metadata but not manifest
    url = langsmith_client.push_prompt(
        prompt_name,
        is_public=False,
        description="Updated prompt",
    )

    updated_prompt = langsmith_client.get_prompt(prompt_name)
    assert isinstance(updated_prompt, ls_schemas.Prompt)
    assert updated_prompt.description == "Updated prompt"
    assert not updated_prompt.is_public
    assert updated_prompt.num_commits == 1

    # test updating prompt manifest but not metadata
    url = langsmith_client.push_prompt(
        prompt_name,
        object=prompt_template_2,
    )
    assert isinstance(url, str)

    langsmith_client.delete_prompt(prompt_name)


@pytest.mark.parametrize("is_public,expected_count", [(True, 1), (False, 1)])
@skip_if_rate_limited
def test_list_prompts_filter(
    langsmith_client: Client,
    prompt_template_1: ChatPromptTemplate,
    is_public: bool,
    expected_count: int,
):
    prompt_name = f"test_list_filter_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(
        prompt_name, object=prompt_template_1, is_public=is_public
    )

    response = langsmith_client.list_prompts(is_public=is_public, query=prompt_name)

    assert response.total == expected_count
    if expected_count > 0:
        assert response.repos[0].repo_handle == prompt_name

    langsmith_client.delete_prompt(prompt_name)


@skip_if_rate_limited
def test_update_prompt_archive(
    langsmith_client: Client, prompt_template_1: ChatPromptTemplate
):
    prompt_name = f"test_archive_{uuid4().hex[:8]}"
    langsmith_client.push_prompt(prompt_name, object=prompt_template_1)

    langsmith_client.update_prompt(prompt_name, is_archived=True)
    archived_prompt = langsmith_client.get_prompt(prompt_name)
    assert isinstance(archived_prompt, ls_schemas.Prompt)
    assert archived_prompt.is_archived

    langsmith_client.update_prompt(prompt_name, is_archived=False)
    unarchived_prompt = langsmith_client.get_prompt(prompt_name)
    assert isinstance(unarchived_prompt, ls_schemas.Prompt)
    assert not unarchived_prompt.is_archived

    langsmith_client.delete_prompt(prompt_name)


@pytest.mark.parametrize(
    "sort_field, sort_direction",
    [
        (ls_schemas.PromptSortField.updated_at, "desc"),
    ],
)
@skip_if_rate_limited
def test_list_prompts_sorting(
    langsmith_client: Client,
    prompt_template_1: ChatPromptTemplate,
    sort_field: ls_schemas.PromptSortField,
    sort_direction: Literal["asc", "desc"],
):
    prompt_names = [f"test_sort_{i}_{uuid4().hex[:8]}" for i in range(3)]
    for name in prompt_names:
        langsmith_client.push_prompt(name, object=prompt_template_1)

    response = langsmith_client.list_prompts(
        sort_field=sort_field, sort_direction=sort_direction, limit=10
    )

    assert len(response.repos) >= 3
    sorted_names = [
        repo.repo_handle for repo in response.repos if repo.repo_handle in prompt_names
    ]
    assert sorted_names == sorted(sorted_names, reverse=(sort_direction == "desc"))

    for name in prompt_names:
        langsmith_client.delete_prompt(name)


def test_convert_to_openai_format(chat_prompt_template: ChatPromptTemplate):
    invoked = chat_prompt_template.invoke({"question": "What is the meaning of life?"})

    res = convert_prompt_to_openai_format(
        invoked,
    )
    expected = {
        "messages": [
            {"content": "You are a chatbot", "role": "system"},
            {"content": "What is the meaning of life?", "role": "user"},
        ],
        "model": "gpt-3.5-turbo",
        "stream": False,
    }
    assert {k: res[k] for k in expected.keys()} == expected


def test_convert_to_anthropic_format(chat_prompt_template: ChatPromptTemplate):
    invoked = chat_prompt_template.invoke({"question": "What is the meaning of life?"})

    res = convert_prompt_to_anthropic_format(invoked, {"model_name": "claude-2"})

    assert res == {
        "model": "claude-2",
        "max_tokens": AnyInt(1024),
        "messages": [{"role": "user", "content": "What is the meaning of life?"}],
        "system": "You are a chatbot",
    }



"""Configuration file for the Sphinx documentation builder."""

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import sys
from pathlib import Path

import toml
from docutils import nodes
from docutils.parsers.rst.directives.admonitions import BaseAdmonition
from docutils.statemachine import StringList
from sphinx.util.docutils import SphinxDirective

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

_DIR = Path(__file__).parent.absolute()
sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../python"))

with (_DIR.parent / "pyproject.toml").open("r") as f:
    data = toml.load(f)


class ExampleLinksDirective(SphinxDirective):
    """Directive to generate a list of links to examples.

    We have a script that extracts links to API reference docs
    from our notebook examples. This directive uses that information
    to backlink to the examples from the API reference docs.
    """

    has_content = False
    required_arguments = 1

    def run(self):
        """Run the directive.

        Called any time :example_links:`ClassName` is used
        in the template *.rst files.
        """
        class_or_func_name = self.arguments[0]
        links = {}
        list_node = nodes.bullet_list()
        for doc_name, link in sorted(links.items()):
            item_node = nodes.list_item()
            para_node = nodes.paragraph()
            link_node = nodes.reference()
            link_node["refuri"] = link
            link_node.append(nodes.Text(doc_name))
            para_node.append(link_node)
            item_node.append(para_node)
            list_node.append(item_node)
        if list_node.children:
            title_node = nodes.rubric()
            title_node.append(nodes.Text(f"Examples using {class_or_func_name}"))
            return [title_node, list_node]
        return [list_node]


class Beta(BaseAdmonition):
    required_arguments = 0
    node_class = nodes.admonition

    def run(self):
        self.content = self.content or StringList(
            [
                (
                    "This feature is in beta. It is actively being worked on, so the "
                    "API may change."
                )
            ]
        )
        self.arguments = self.arguments or ["Beta"]
        return super().run()


def setup(app):
    app.add_directive("example_links", ExampleLinksDirective)
    app.add_directive("beta", Beta)


# -- Project information -----------------------------------------------------

project = "🦜️🛠️ LangSmith"
copyright = "2024, LangChain Inc"
author = "LangChain, Inc"

html_favicon = "_static/img/brand/favicon.png"
html_last_updated_fmt = "%b %d, %Y"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autodoc.typehints",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxcontrib.autodoc_pydantic",
    # "IPython.sphinxext.ipython_console_highlighting",
    "myst_parser",
    "_extensions.gallery_directive",
    "sphinx_design",
    "sphinx_copybutton",
]
source_suffix = [".rst", ".md"]

# some autodoc pydantic options are repeated in the actual template.
# potentially user error, but there may be bugs in the sphinx extension
# with options not being passed through correctly (from either the location in the code)
autodoc_pydantic_model_show_json = False
autodoc_pydantic_field_list_validators = False
autodoc_pydantic_config_members = False
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_model_show_validator_members = False
autodoc_pydantic_model_show_validator_summary = False
autodoc_pydantic_model_signature_prefix = "class"
autodoc_pydantic_field_signature_prefix = "param"
autodoc_member_order = "groupwise"
autoclass_content = "both"
autodoc_typehints_format = "short"
autodoc_typehints = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# The theme to use for HTML and HTML Help pages.
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    #     # -- General configuration ------------------------------------------------
    "sidebar_includehidden": True,
    "use_edit_page_button": False,
    #     # "analytics": {
    #     #     "plausible_analytics_domain": "scikit-learn.org",
    #     #     "plausible_analytics_url": "https://views.scientific-python.org/js/script.js",
    #     # },
    #     # If "prev-next" is included in article_footer_items, then setting show_prev_next
    #     # to True would repeat prev and next links. See
    #     # https://github.com/pydata/pydata-sphinx-theme/blob/b731dc230bc26a3d1d1bb039c56c977a9b3d25d8/src/pydata_sphinx_theme/theme/pydata_sphinx_theme/layout.html#L118-L129
    "show_prev_next": False,
    "search_bar_text": "Search",
    "navigation_with_keys": True,
    "collapse_navigation": True,
    "navigation_depth": 3,
    "show_nav_level": 1,
    "show_toc_level": 3,
    "navbar_align": "left",
    "header_links_before_dropdown": 5,
    "header_dropdown_text": "Modules",
    "logo": {
        "image_light": "_static/wordmark-api.svg",
        "image_dark": "_static/wordmark-api-dark.svg",
    },
    "surface_warnings": True,
    #     # -- Template placement in theme layouts ----------------------------------
    "navbar_start": ["navbar-logo"],
    #     # Note that the alignment of navbar_center is controlled by navbar_align
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["langsmith_docs", "theme-switcher", "navbar-icon-links"],
    #     # navbar_persistent is persistent right (even when on mobiles)
    "navbar_persistent": ["search-field"],
    "article_header_start": ["breadcrumbs"],
    "article_header_end": [],
    "article_footer_items": [],
    "content_footer_items": [],
    #     # Use html_sidebars that map page patterns to list of sidebar templates
    #     "primary_sidebar_end": [],
    "footer_start": ["copyright"],
    "footer_center": [],
    "footer_end": [],
    #     # When specified as a dictionary, the keys should follow glob-style patterns, as in
    #     # https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-exclude_patterns
    #     # In particular, "**" specifies the default for all pages
    #     # Use :html_theme.sidebar_secondary.remove: for file-wide removal
    #     "secondary_sidebar_items": {"**": ["page-toc", "sourcelink"]},
    #     "show_version_warning_banner": True,
    #     "announcement": None,
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/langchain-ai/langsmith-sdk",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fa-brands fa-square-github",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        },
        {
            "name": "X / Twitter",
            "url": "https://twitter.com/langchainai",
            "icon": "fab fa-twitter-square",
        },
    ],
    "icon_links_label": "Quick Links",
    "external_links": [],
    "sidebar_primary_title": "Your Custom Title Here",
}


html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "langchain-ai",  # Username
    "github_repo": "langsmith-sdk",  # Repo name
    "github_version": "master",  # Version
    "conf_py_path": "/docs/api_reference",  # Path in the checkout to the docs root
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (e.g. https://...)
html_css_files = ["css/custom.css"]
html_use_index = False

myst_enable_extensions = ["colon_fence"]

# generate autosummary even if no references
autosummary_generate = True

html_copy_source = False
html_show_sourcelink = False

# Set canonical URL from the Read the Docs Domain
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")

# Tell Jinja2 templates the build is running on Read the Docs
if os.environ.get("READTHEDOCS", "") == "True":
    html_context["READTHEDOCS"] = True

master_doc = "index"
templates_path = ["templates"]

# If a signature’s length in characters exceeds this threshold,
# each parameter within the signature will be displayed on an individual logical line
maximum_signature_line_length = 60



"""Tests for sandbox data models."""

from langsmith.sandbox import (
    ExecutionResult,
    OutputChunk,
    Pool,
    ResourceSpec,
    ResourceStatus,
    SandboxTemplate,
    Volume,
    VolumeMountSpec,
)


class TestOutputChunk:
    """Tests for OutputChunk."""

    def test_dataclass(self):
        """Test OutputChunk fields."""
        chunk = OutputChunk(stream="stdout", data="hello", offset=0)
        assert chunk.stream == "stdout"
        assert chunk.data == "hello"
        assert chunk.offset == 0


class TestExecutionResult:
    """Tests for ExecutionResult."""

    def test_success_property_true(self):
        """Test success is True when exit_code is 0."""
        result = ExecutionResult(stdout="output", stderr="", exit_code=0)
        assert result.success is True

    def test_success_property_false(self):
        """Test success is False when exit_code is non-zero."""
        result = ExecutionResult(stdout="", stderr="error", exit_code=1)
        assert result.success is False

    def test_success_property_negative_exit(self):
        """Test success is False when exit_code is negative."""
        result = ExecutionResult(stdout="", stderr="", exit_code=-1)
        assert result.success is False


class TestResourceSpec:
    """Tests for ResourceSpec."""

    def test_default_values(self):
        """Test default values."""
        spec = ResourceSpec()
        assert spec.cpu == "500m"
        assert spec.memory == "512Mi"
        assert spec.storage is None

    def test_custom_values(self):
        """Test custom values."""
        spec = ResourceSpec(cpu="1", memory="1Gi", storage="5Gi")
        assert spec.cpu == "1"
        assert spec.memory == "1Gi"
        assert spec.storage == "5Gi"


class TestVolume:
    """Tests for Volume."""

    def test_from_dict(self):
        """Test creating from dict."""
        data = {
            "id": "550e8400-e29b-41d4-a716-446655440000",
            "name": "test-volume",
            "size": "1Gi",
            "storage_class": "hostpath",
            "created_at": "2025-01-01T00:00:00Z",
            "updated_at": "2025-01-02T00:00:00Z",
        }
        volume = Volume.from_dict(data)

        assert volume.id == "550e8400-e29b-41d4-a716-446655440000"
        assert volume.name == "test-volume"
        assert volume.size == "1Gi"
        assert volume.storage_class == "hostpath"
        assert volume.created_at == "2025-01-01T00:00:00Z"
        assert volume.updated_at == "2025-01-02T00:00:00Z"

    def test_from_dict_minimal(self):
        """Test creating from minimal dict."""
        data = {
            "name": "test-volume",
            "size": "5Gi",
        }
        volume = Volume.from_dict(data)

        assert volume.id is None
        assert volume.name == "test-volume"
        assert volume.size == "5Gi"
        assert volume.storage_class == "default"
        assert volume.created_at is None
        assert volume.updated_at is None


class TestVolumeMountSpec:
    """Tests for VolumeMountSpec."""

    def test_creation(self):
        """Test creating a volume mount spec."""
        mount = VolumeMountSpec(volume_name="my-volume", mount_path="/data")
        assert mount.volume_name == "my-volume"
        assert mount.mount_path == "/data"


class TestSandboxTemplate:
    """Tests for SandboxTemplate."""

    def test_from_dict_minimal(self):
        """Test creating from minimal dict."""
        data = {
            "name": "test-template",
            "image": "python:3.12",
        }
        template = SandboxTemplate.from_dict(data)

        assert template.id is None
        assert template.name == "test-template"
        assert template.image == "python:3.12"
        assert template.resources.cpu == "500m"
        assert template.updated_at is None

    def test_from_dict_full(self):
        """Test creating from full dict."""
        data = {
            "id": "550e8400-e29b-41d4-a716-446655440001",
            "name": "test-template",
            "image": "node:20",
            "resources": {
                "cpu": "2",
                "memory": "4Gi",
                "storage": "10Gi",
            },
            "created_at": "2025-01-01T00:00:00Z",
            "updated_at": "2025-01-02T00:00:00Z",
        }
        template = SandboxTemplate.from_dict(data)

        assert template.id == "550e8400-e29b-41d4-a716-446655440001"
        assert template.name == "test-template"
        assert template.image == "node:20"
        assert template.resources.cpu == "2"
        assert template.resources.memory == "4Gi"
        assert template.resources.storage == "10Gi"
        assert template.created_at == "2025-01-01T00:00:00Z"
        assert template.updated_at == "2025-01-02T00:00:00Z"

    def test_from_dict_with_volume_mounts(self):
        """Test creating from dict with volume mounts."""
        data = {
            "name": "test-template",
            "image": "python:3.12",
            "volume_mounts": [
                {"volume_name": "data-volume", "mount_path": "/data"},
                {"volume_name": "cache-volume", "mount_path": "/cache"},
            ],
        }
        template = SandboxTemplate.from_dict(data)

        assert len(template.volume_mounts) == 2
        assert template.volume_mounts[0].volume_name == "data-volume"
        assert template.volume_mounts[0].mount_path == "/data"
        assert template.volume_mounts[1].volume_name == "cache-volume"
        assert template.volume_mounts[1].mount_path == "/cache"

    def test_from_dict_empty_volume_mounts(self):
        """Test creating from dict with empty volume mounts."""
        data = {
            "name": "test-template",
            "image": "python:3.12",
            "volume_mounts": [],
        }
        template = SandboxTemplate.from_dict(data)

        assert template.volume_mounts == []

    def test_from_dict_no_volume_mounts_key(self):
        """Test creating from dict without volume_mounts key."""
        data = {
            "name": "test-template",
            "image": "python:3.12",
        }
        template = SandboxTemplate.from_dict(data)

        assert template.volume_mounts == []


class TestResourceStatus:
    """Tests for ResourceStatus."""

    def test_from_dict_provisioning(self):
        """Test creating from provisioning response."""
        data = {"status": "provisioning", "status_message": None}
        status = ResourceStatus.from_dict(data)
        assert status.status == "provisioning"
        assert status.status_message is None

    def test_from_dict_ready(self):
        """Test creating from ready response."""
        data = {"status": "ready", "status_message": None}
        status = ResourceStatus.from_dict(data)
        assert status.status == "ready"
        assert status.status_message is None

    def test_from_dict_failed(self):
        """Test creating from failed response with message."""
        data = {
            "status": "failed",
            "status_message": "No capacity available",
        }
        status = ResourceStatus.from_dict(data)
        assert status.status == "failed"
        assert status.status_message == "No capacity available"

    def test_from_dict_defaults(self):
        """Test defaults when keys are missing."""
        status = ResourceStatus.from_dict({})
        assert status.status == "provisioning"
        assert status.status_message is None


class TestPool:
    """Tests for Pool."""

    def test_from_dict_full(self):
        """Test creating from full dict."""
        data = {
            "id": "550e8400-e29b-41d4-a716-446655440002",
            "name": "python-pool",
            "template_name": "python-sandbox",
            "replicas": 5,
            "created_at": "2026-01-16T12:00:00Z",
            "updated_at": "2026-01-16T14:30:00Z",
        }
        pool = Pool.from_dict(data)

        assert pool.id == "550e8400-e29b-41d4-a716-446655440002"
        assert pool.name == "python-pool"
        assert pool.template_name == "python-sandbox"
        assert pool.replicas == 5
        assert pool.created_at == "2026-01-16T12:00:00Z"
        assert pool.updated_at == "2026-01-16T14:30:00Z"

    def test_from_dict_minimal(self):
        """Test creating from minimal dict."""
        data = {
            "name": "python-pool",
            "template_name": "python-sandbox",
            "replicas": 3,
        }
        pool = Pool.from_dict(data)

        assert pool.id is None
        assert pool.name == "python-pool"
        assert pool.template_name == "python-sandbox"
        assert pool.replicas == 3
        assert pool.created_at is None
        assert pool.updated_at is None

    def test_from_dict_paused(self):
        """Test creating from dict for paused pool."""
        data = {
            "name": "paused-pool",
            "template_name": "python-sandbox",
            "replicas": 0,
        }
        pool = Pool.from_dict(data)

        assert pool.replicas == 0



"""Test custom headers functionality."""

from unittest import mock

import pytest

from langsmith import Client


def _clear_env_cache():
    """Clear the environment variable cache."""
    try:
        from langsmith import env

        env._runtime_env.cache_clear()
    except Exception:
        pass


@mock.patch("langsmith.client.requests.Session")
def test_custom_headers(
    mock_session: mock.Mock, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test that custom headers are properly merged with default headers."""
    _clear_env_cache()
    monkeypatch.delenv("LANGCHAIN_API_KEY", raising=False)

    with mock.patch.dict("os.environ", {}, clear=True):
        # Test with custom headers
        custom_headers = {
            "X-Custom-Header": "custom-value",
            "X-Another-Header": "another-value",
        }

        client = Client(
            api_url="http://localhost:1984",
            api_key="test-api-key",
            headers=custom_headers,
        )

        # Check that custom headers are included
        assert "X-Custom-Header" in client._headers
        assert client._headers["X-Custom-Header"] == "custom-value"
        assert "X-Another-Header" in client._headers
        assert client._headers["X-Another-Header"] == "another-value"

        # Check that default headers are still present
        assert "User-Agent" in client._headers
        assert "Accept" in client._headers
        assert "x-api-key" in client._headers
        assert client._headers["x-api-key"] == "test-api-key"


@mock.patch("langsmith.client.requests.Session")
def test_custom_headers_dont_override_required(
    mock_session: mock.Mock, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test that custom headers don't override required headers."""
    _clear_env_cache()
    monkeypatch.delenv("LANGCHAIN_API_KEY", raising=False)

    with mock.patch.dict("os.environ", {}, clear=True):
        # Try to override the API key with custom headers (should not work)
        custom_headers = {
            "x-api-key": "wrong-key",
            "X-Custom-Header": "custom-value",
        }

        client = Client(
            api_url="http://localhost:1984",
            api_key="correct-api-key",
            headers=custom_headers,
        )

        # Check that the API key from the parameter is used, not the custom header
        assert client._headers["x-api-key"] == "correct-api-key"
        # But the custom header should still be present
        assert client._headers["X-Custom-Header"] == "custom-value"


@mock.patch("langsmith.client.requests.Session")
def test_no_custom_headers(
    mock_session: mock.Mock, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test that client works without custom headers."""
    _clear_env_cache()
    monkeypatch.delenv("LANGCHAIN_API_KEY", raising=False)

    with mock.patch.dict("os.environ", {}, clear=True):
        client = Client(
            api_url="http://localhost:1984",
            api_key="test-api-key",
        )

        # Check that default headers are present
        assert "User-Agent" in client._headers
        assert "Accept" in client._headers
        assert "x-api-key" in client._headers
        assert client._headers["x-api-key"] == "test-api-key"


@mock.patch("langsmith.client.requests.Session")
def test_workspace_id_with_custom_headers(
    mock_session: mock.Mock, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test that workspace_id header is preserved with custom headers."""
    _clear_env_cache()
    monkeypatch.delenv("LANGCHAIN_API_KEY", raising=False)

    with mock.patch.dict("os.environ", {}, clear=True):
        custom_headers = {
            "X-Custom-Header": "custom-value",
        }

        client = Client(
            api_url="http://localhost:1984",
            api_key="test-api-key",
            workspace_id="test-workspace",
            headers=custom_headers,
        )

        # Check that workspace ID header is present
        assert "X-Tenant-Id" in client._headers
        assert client._headers["X-Tenant-Id"] == "test-workspace"
        # And custom header is also present
        assert client._headers["X-Custom-Header"] == "custom-value"



import statistics
import time
import zlib
from concurrent.futures import ThreadPoolExecutor

import orjson


def create_json_with_large_array(length):
    """Create a large JSON object for benchmarking purposes."""
    large_array = [
        {
            "index": i,
            "data": f"This is element number {i}",
            "nested": {"id": i, "value": f"Nested value for element {i}"},
        }
        for i in range(length)
    ]

    return {
        "name": "Huge JSON",
        "description": "This is a very large JSON object for benchmarking purposes.",
        "array": large_array,
        "metadata": {
            "created_at": "2024-10-22T19:00:00Z",
            "author": "Python Program",
            "version": 1.0,
        },
    }


def create_json_with_large_strings(length: int) -> dict:
    large_string = "a" * length  # Create a large string of repeated 'a' characters

    return {
        "name": "Huge JSON",
        "description": "This is a very large JSON object for benchmarking purposes.",
        "key1": large_string,
        "key2": large_string,
        "key3": large_string,
        "metadata": {
            "created_at": "2024-10-22T19:00:00Z",
            "author": "Python Program",
            "version": 1.0,
        },
    }


def serialize_sequential(data):
    """Serialize data sequentially."""
    return [orjson.dumps(json_obj) for json_obj in data]


def serialize_parallel(data):
    """Serialize data in parallel using ThreadPoolExecutor."""
    with ThreadPoolExecutor() as executor:
        return list(executor.map(orjson.dumps, data))


def serialize_sequential_gz(data):
    """Serialize data sequentially and compress using zlib.

    With adjustable compression level."""
    compressed_data = []
    for json_obj in data:
        serialized = orjson.dumps(json_obj)
        compressed = zlib.compress(serialized, level=1)
        compressed_data.append(compressed)
    return compressed_data


def serialize_parallel_gz(data):
    """Serialize data in parallel with zlib.

    Using ThreadPoolExecutor and zlib with adjustable compression level."""

    def compress_item(json_obj):
        serialized = orjson.dumps(json_obj)
        return zlib.compress(serialized, level=1)

    with ThreadPoolExecutor() as executor:
        compressed_data = list(executor.map(compress_item, data))
    return compressed_data


def gzip_parallel(serialized_data):
    """Compress serialized data in parallel using ThreadPoolExecutor and zlib."""
    with ThreadPoolExecutor() as executor:
        return list(executor.map(zlib.compress, serialized_data))


def gzip_sequential(serialized_data):
    """Compress serialized data sequentially using zlib."""
    return [zlib.compress(serialized) for serialized in serialized_data]


def benchmark_serialization(data, func, samples=10):
    """Benchmark a serialization function with multiple samples."""
    timings = []
    for _ in range(samples):
        start = time.perf_counter()
        func(data)
        elapsed = time.perf_counter() - start
        timings.append(elapsed)

    return {
        "mean": statistics.mean(timings),
        "median": statistics.median(timings),
        "stdev": statistics.stdev(timings) if len(timings) > 1 else 0,
        "min": min(timings),
        "max": max(timings),
    }


def main():
    num_json_objects = 2000
    json_length = 5000

    data = [create_json_with_large_array(json_length) for _ in range(num_json_objects)]
    serialized_data = serialize_sequential(data)

    for func in [
        serialize_sequential,
        serialize_parallel,
        serialize_sequential_gz,
        serialize_parallel_gz,
        gzip_sequential,
        gzip_parallel,
    ]:
        # data = [
        #     create_json_with_large_strings(json_length)
        #     for _ in range(num_json_objects)
        # ]

        print(
            f"\nBenchmarking {func.__name__} with {num_json_objects} JSON objects "
            f"of length {json_length}..."
        )
        results_seq = (
            benchmark_serialization(data, func)
            if not func.__name__.startswith("gzip")
            else benchmark_serialization(serialized_data, func)
        )
        print(f"Mean time: {results_seq['mean']:.4f} seconds")
        print(f"Median time: {results_seq['median']:.4f} seconds")
        print(f"Std Dev: {results_seq['stdev']:.4f} seconds")
        print(f"Min time: {results_seq['min']:.4f} seconds")
        print(f"Max time: {results_seq['max']:.4f} seconds")


if __name__ == "__main__":
    main()



"""
Unit test that mocks LANGSMITH_RUNS_ENDPOINTS and verifies that multiple requests
are made to the respective endpoints for the same run.

This test covers:
1. Direct client.create_run() calls with api_key/api_url parameters
2. RunTree.post() with replicas that have api_key/api_url
3. Background threading mode with different endpoints
4. Both multipart and non-multipart modes
"""

import json
import uuid
from unittest.mock import Mock, patch

import pytest

from langsmith import Client
from langsmith._internal._background_thread import (
    TracingQueueItem,
    _tracing_thread_handle_batch,
)
from langsmith._internal._operations import serialize_run_dict
from langsmith.run_trees import ApiKeyAuth, RunTree, WriteReplica


class TestLangsmithRunsEndpoints:
    """Test LANGSMITH_RUNS_ENDPOINTS functionality with multiple endpoints."""

    def setup_method(self):
        """Set up test data."""
        self.run_data = {
            "id": uuid.uuid4(),
            "trace_id": uuid.uuid4(),
            "dotted_order": "20231201T120000000000Z" + str(uuid.uuid4()),
            "session_name": "test-project",
            "name": "test_run",
            "inputs": {"input": "test"},
            "run_type": "llm",
        }

        self.endpoints_config = [
            {"api_url": "https://api1.example.com", "api_key": "key1"},
            {"api_url": "https://api2.example.com", "api_key": "key2"},
            {"api_url": "https://api3.example.com", "api_key": "key3"},
        ]

    def test_client_create_run_with_multiple_write_api_urls(self):
        """Test that client.create_run() sends to multiple endpoints when using
        _write_api_urls.
        """
        # Create client with multiple write API URLs
        api_urls = {
            "https://api1.example.com": "key1",
            "https://api2.example.com": "key2",
            "https://api3.example.com": "key3",
        }

        client = Client(api_urls=api_urls, auto_batch_tracing=False)

        with patch.object(client.session, "request") as mock_request:
            mock_response = Mock()
            mock_response.json.return_value = {}
            mock_request.return_value = mock_response

            # Call create_run without specific api_key/api_url
            # (should use all endpoints)
            client.create_run(**self.run_data)

            # Verify requests were made to all 3 endpoints
            assert mock_request.call_count == 3

            # Extract the URLs from the calls
            called_urls = []
            for call_args in mock_request.call_args_list:
                # The URL is the second positional argument:
                # session.request(method, url, ...)
                full_url = call_args[0][1]  # call_args[0] is args tuple, [1] is the URL
                base_url = full_url.replace("/runs", "")
                called_urls.append(base_url)

            # Verify all endpoints were called
            assert "https://api1.example.com" in called_urls
            assert "https://api2.example.com" in called_urls
            assert "https://api3.example.com" in called_urls

            # Verify correct API keys were used
            for i, call_args in enumerate(mock_request.call_args_list):
                # Headers are in kwargs
                headers = call_args[1].get("headers", {})
                api_key = headers.get("x-api-key")  # Note: lowercase header key
                assert api_key in ["key1", "key2", "key3"]

    def test_client_create_run_with_specific_endpoint(self):
        """Test that client.create_run() with specific api_key/api_url only sends
        to that endpoint.
        """
        api_urls = {
            "https://api1.example.com": "key1",
            "https://api2.example.com": "key2",
        }

        client = Client(api_urls=api_urls, auto_batch_tracing=False)

        with patch.object(client.session, "request") as mock_request:
            mock_response = Mock()
            mock_response.json.return_value = {}
            mock_request.return_value = mock_response

            # Call create_run with specific endpoint
            client.create_run(
                **self.run_data,
                api_key="custom_key",
                api_url="https://custom.example.com",
            )

            # Verify only one request was made to the specific endpoint
            assert mock_request.call_count == 1
            call_args = mock_request.call_args
            full_url = call_args[0][1]  # Second positional argument is the URL
            assert full_url == "https://custom.example.com/runs"

            headers = call_args[1].get("headers", {})
            # Note: lowercase header key
            assert headers.get("x-api-key") == "custom_key"

    def test_run_tree_with_replicas(self):
        """Test RunTree.post() with replicas containing auth/api_url."""
        client = Mock()

        # Create RunTree with replicas that have different endpoints
        replicas = [
            WriteReplica(
                api_url="https://replica1.example.com",
                auth=ApiKeyAuth(api_key="replica_key1"),
                project_name="project1",
            ),
            WriteReplica(
                api_url="https://replica2.example.com",
                auth=ApiKeyAuth(api_key="replica_key2"),
                project_name="project2",
            ),
        ]

        run_tree = RunTree(
            name="test_run",
            inputs={"input": "test"},
            client=client,
            project_name="main-project",
            replicas=replicas,
        )

        # Call post()
        run_tree.post()

        # Verify client.create_run was called once for each replica
        assert client.create_run.call_count == 2

        # Verify the calls had correct api_key/api_url parameters
        calls = client.create_run.call_args_list

        # First replica call
        first_call = calls[0]
        assert first_call.kwargs["api_key"] == "replica_key1"
        assert first_call.kwargs["api_url"] == "https://replica1.example.com"
        assert first_call.kwargs["session_name"] == "project1"

        # Second replica call
        second_call = calls[1]
        assert second_call.kwargs["api_key"] == "replica_key2"
        assert second_call.kwargs["api_url"] == "https://replica2.example.com"
        assert second_call.kwargs["session_name"] == "project2"

    def test_background_threading_with_different_endpoints(self):
        """Test background threading correctly groups and sends to different
        endpoints.
        """
        client = Mock()
        tracing_queue = Mock()

        # Mock the client methods
        client._multipart_ingest_ops = Mock()
        client._batch_ingest_run_ops = Mock()

        # Create batch with items for different endpoints
        serialized_op1 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op2 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op3 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )

        batch = [
            TracingQueueItem(
                "priority1",
                serialized_op1,
                api_key="key1",
                api_url="https://api1.com",
            ),
            # Same endpoint
            TracingQueueItem(
                "priority2",
                serialized_op2,
                api_key="key1",
                api_url="https://api1.com",
            ),
            # Different endpoint
            TracingQueueItem(
                "priority3",
                serialized_op3,
                api_key="key2",
                api_url="https://api2.com",
            ),
        ]

        # Test multipart mode
        _tracing_thread_handle_batch(client, tracing_queue, batch, use_multipart=True)

        # Verify _multipart_ingest_ops was called twice (once per unique endpoint)
        assert client._multipart_ingest_ops.call_count == 2

        # Verify the calls had correct endpoint parameters
        calls = client._multipart_ingest_ops.call_args_list
        endpoint_calls = [
            (call.kwargs.get("api_url"), call.kwargs.get("api_key")) for call in calls
        ]

        assert ("https://api1.com", "key1") in endpoint_calls
        assert ("https://api2.com", "key2") in endpoint_calls

        # Verify the first endpoint got 2 operations, second got 1
        for call in calls:
            ops = call.args[0]  # First positional argument is the operations list
            if call.kwargs.get("api_url") == "https://api1.com":
                assert len(ops) == 2  # Two operations for api1
            else:
                assert len(ops) == 1  # One operation for api2

    def test_background_threading_groups_by_service_auth(self):
        """Test background threading groups by (api_url, service_key, tenant_id)."""
        client = Mock()
        tracing_queue = Mock()

        client._multipart_ingest_ops = Mock()

        serialized_op1 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op2 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )

        batch = [
            TracingQueueItem(
                "priority1",
                serialized_op1,
                api_url="https://api.same.com",
                service_key="jwt-a",
                tenant_id="tenant-a",
            ),
            TracingQueueItem(
                "priority2",
                serialized_op2,
                api_url="https://api.same.com",
                service_key="jwt-b",
                tenant_id="tenant-b",
            ),
        ]

        _tracing_thread_handle_batch(client, tracing_queue, batch, use_multipart=True)

        assert client._multipart_ingest_ops.call_count == 2
        calls = client._multipart_ingest_ops.call_args_list
        auth_calls = [
            (
                call.kwargs.get("api_url"),
                call.kwargs.get("service_key"),
                call.kwargs.get("tenant_id"),
            )
            for call in calls
        ]
        assert ("https://api.same.com", "jwt-a", "tenant-a") in auth_calls
        assert ("https://api.same.com", "jwt-b", "tenant-b") in auth_calls

    def test_background_threading_groups_by_auth_headers(self):
        """Test background threading groups by authorization/cookie overrides."""
        client = Mock()
        tracing_queue = Mock()

        client._multipart_ingest_ops = Mock()

        serialized_op1 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op2 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )

        batch = [
            TracingQueueItem(
                "priority1",
                serialized_op1,
                api_url="https://api.same.com",
                api_key="shared-key",
                authorization="Bearer token-a",
                cookie="session=a",
            ),
            TracingQueueItem(
                "priority2",
                serialized_op2,
                api_url="https://api.same.com",
                api_key="shared-key",
                authorization="Bearer token-b",
                cookie="session=b",
            ),
        ]

        _tracing_thread_handle_batch(client, tracing_queue, batch, use_multipart=True)

        assert client._multipart_ingest_ops.call_count == 2
        calls = client._multipart_ingest_ops.call_args_list
        auth_pairs = {
            (call.kwargs.get("authorization"), call.kwargs.get("cookie"))
            for call in calls
        }
        assert ("Bearer token-a", "session=a") in auth_pairs
        assert ("Bearer token-b", "session=b") in auth_pairs
        assert all(len(call.args[0]) == 1 for call in calls)

    def test_background_threading_does_not_mix_service_and_api_key_auth(self):
        """Service auth and api key auth should never be grouped together."""
        client = Mock()
        tracing_queue = Mock()

        client._multipart_ingest_ops = Mock()

        serialized_op1 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op2 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )

        batch = [
            TracingQueueItem(
                "priority1",
                serialized_op1,
                api_url="https://api.same.com",
                api_key="key-a",
            ),
            TracingQueueItem(
                "priority2",
                serialized_op2,
                api_url="https://api.same.com",
                service_key="jwt-b",
                tenant_id="tenant-b",
            ),
        ]

        _tracing_thread_handle_batch(client, tracing_queue, batch, use_multipart=True)

        assert client._multipart_ingest_ops.call_count == 2

    def test_background_threading_non_multipart_mode(self):
        """Test background threading in non-multipart mode with different endpoints."""
        client = Mock()
        tracing_queue = Mock()

        # Mock the client methods
        client._batch_ingest_run_ops = Mock()

        # Create batch with items for different endpoints
        serialized_op1 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op2 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )

        batch = [
            TracingQueueItem(
                "priority1",
                serialized_op1,
                api_key="key1",
                api_url="https://api1.com",
            ),
            TracingQueueItem(
                "priority2",
                serialized_op2,
                api_key="key2",
                api_url="https://api2.com",
            ),
        ]

        # Test non-multipart mode
        _tracing_thread_handle_batch(client, tracing_queue, batch, use_multipart=False)

        # Verify _batch_ingest_run_ops was called twice (once per endpoint)
        assert client._batch_ingest_run_ops.call_count == 2

        # Verify the calls had correct endpoint parameters
        calls = client._batch_ingest_run_ops.call_args_list
        endpoint_calls = [
            (call.kwargs.get("api_url"), call.kwargs.get("api_key")) for call in calls
        ]

        assert ("https://api1.com", "key1") in endpoint_calls
        assert ("https://api2.com", "key2") in endpoint_calls

    def test_langsmith_runs_endpoints_env_var_integration(self):
        """Test integration with LANGSMITH_RUNS_ENDPOINTS environment variable."""
        from langsmith.run_trees import _parse_write_replicas_from_env_var

        # Test the parsing function directly
        env_var = json.dumps(self.endpoints_config)
        replicas = _parse_write_replicas_from_env_var(env_var)

        assert len(replicas) == 3
        assert replicas[0]["api_url"] == "https://api1.example.com"
        assert replicas[0]["auth"]["api_key"] == "key1"
        assert replicas[1]["api_url"] == "https://api2.example.com"
        assert replicas[1]["auth"]["api_key"] == "key2"
        assert replicas[2]["api_url"] == "https://api3.example.com"
        assert replicas[2]["auth"]["api_key"] == "key3"

    def test_mixed_endpoints_and_default_fallback(self):
        """Test batch with mixed endpoints and items that should use default."""
        client = Mock()
        tracing_queue = Mock()
        client._multipart_ingest_ops = Mock()

        # Create batch with mixed endpoint specifications
        serialized_op1 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op2 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )
        serialized_op3 = serialize_run_dict(
            "post", {**self.run_data, "id": uuid.uuid4()}
        )

        batch = [
            TracingQueueItem(
                "priority1",
                serialized_op1,
                api_key="key1",
                api_url="https://api1.com",
            ),
            # Should use default
            TracingQueueItem(
                "priority2",
                serialized_op2,
                api_key=None,
                api_url=None,
            ),
            TracingQueueItem(
                "priority3",
                serialized_op3,
                api_key="key2",
                api_url="https://api2.com",
            ),
        ]

        _tracing_thread_handle_batch(client, tracing_queue, batch, use_multipart=True)

        # Should have 3 calls: one for each unique endpoint combination
        assert client._multipart_ingest_ops.call_count == 3

        # Verify endpoint combinations
        calls = client._multipart_ingest_ops.call_args_list
        endpoint_calls = [
            (call.kwargs.get("api_url"), call.kwargs.get("api_key")) for call in calls
        ]

        assert ("https://api1.com", "key1") in endpoint_calls
        assert ("https://api2.com", "key2") in endpoint_calls
        assert (None, None) in endpoint_calls  # Default endpoint


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import asyncio
import functools
import logging
import time
from contextlib import contextmanager
from typing import Callable, Sequence, Tuple, TypeVar

import pytest

from langsmith import Client, aevaluate, evaluate, expect, test
from langsmith.evaluation import EvaluationResult, EvaluationResults
from langsmith.schemas import Example, Run

T = TypeVar("T")


@contextmanager
def suppress_warnings():
    logger = logging.getLogger()
    current_level = logger.level
    logger.setLevel(logging.CRITICAL)
    try:
        yield
    finally:
        logger.setLevel(current_level)


def wait_for(
    condition: Callable[[], Tuple[T, bool]],
    max_sleep_time: int = 120,
    sleep_time: int = 3,
) -> T:
    """Wait for a condition to be true."""
    start_time = time.time()
    last_e = None
    while (time.time() - start_time) < max_sleep_time:
        try:
            res, cond = condition()
            if cond:
                return res
        except Exception as e:
            last_e = e
            time.sleep(sleep_time)
    total_time = time.time() - start_time
    if last_e is not None:
        raise last_e
    raise ValueError(f"Callable did not return within {total_time}")


async def test_error_handling_evaluators():
    client = Client()
    _ = client.clone_public_dataset(
        "https://smith.langchain.com/public/419dcab2-1d66-4b94-8901-0357ead390df/d"
    )
    dataset_name = "Evaluate Examples"

    # Case 1: Normal dictionary return
    def error_dict_evaluator(run: Run, example: Example):
        if True:  # This condition ensures the error is always raised
            raise ValueError("Error in dict evaluator")
        return {"key": "dict_key", "score": 1}

    # Case 2: EvaluationResult return
    def error_evaluation_result(run: Run, example: Example):
        if True:  # This condition ensures the error is always raised
            raise ValueError("Error in EvaluationResult evaluator")
        return EvaluationResult(key="eval_result_key", score=1)

    # Case 3: EvaluationResults return
    def error_evaluation_results(run: Run, example: Example):
        if True:  # This condition ensures the error is always raised
            raise ValueError("Error in EvaluationResults evaluator")
        return EvaluationResults(
            results=[
                EvaluationResult(key="eval_results_key1", score=1),
                EvaluationResult(key="eval_results_key2", score=2),
            ]
        )

    # Case 4: Dictionary without 'key' field
    def error_dict_no_key(run: Run, example: Example):
        if True:  # This condition ensures the error is always raised
            raise ValueError("Error in dict without key evaluator")
        return {"score": 1}

    # Case 5: dict-style results
    def error_evaluation_results_dict(run: Run, example: Example):
        if True:  # This condition ensures the error is always raised
            raise ValueError("Error in EvaluationResults dict evaluator")

        return {
            "results": [
                dict(key="eval_results_dict_key1", score=1),
                {"key": "eval_results_dict_key2", "score": 2},
                EvaluationResult(key="eval_results_dict_key3", score=3),
            ]
        }

    def predict(inputs: dict) -> dict:
        return {"output": "Yes"}

    with suppress_warnings():
        sync_results = evaluate(
            predict,
            data=client.list_examples(
                dataset_name=dataset_name,
                as_of="test_version",
            ),
            evaluators=[
                error_dict_evaluator,
                error_evaluation_result,
                error_evaluation_results,
                error_dict_no_key,
                error_evaluation_results_dict,
            ],
            max_concurrency=1,  # To ensure deterministic order
        )

    assert len(sync_results) == 10  # Assuming 10 examples in the dataset

    def check_results(results):
        for result in results:
            eval_results = result["evaluation_results"]["results"]
            assert len(eval_results) == 8

            # Check error handling for each evaluator
            assert eval_results[0].key == "dict_key"
            assert "Error in dict evaluator" in eval_results[0].comment
            assert eval_results[0].extra.get("error") is True

            assert eval_results[1].key == "eval_result_key"
            assert "Error in EvaluationResult evaluator" in eval_results[1].comment
            assert eval_results[1].extra.get("error") is True

            assert eval_results[2].key == "eval_results_key1"
            assert "Error in EvaluationResults evaluator" in eval_results[2].comment
            assert eval_results[2].extra.get("error") is True

            assert eval_results[3].key == "eval_results_key2"
            assert "Error in EvaluationResults evaluator" in eval_results[3].comment
            assert eval_results[3].extra.get("error") is True

            assert eval_results[4].key == "error_dict_no_key"
            assert "Error in dict without key evaluator" in eval_results[4].comment
            assert eval_results[4].extra.get("error") is True

            assert eval_results[5].key == "eval_results_dict_key1"
            assert (
                "Error in EvaluationResults dict evaluator" in eval_results[5].comment
            )
            assert eval_results[5].extra.get("error") is True

            assert eval_results[6].key == "eval_results_dict_key2"
            assert (
                "Error in EvaluationResults dict evaluator" in eval_results[6].comment
            )
            assert eval_results[6].extra.get("error") is True

            assert eval_results[7].key == "eval_results_dict_key3"
            assert (
                "Error in EvaluationResults dict evaluator" in eval_results[7].comment
            )
            assert eval_results[7].extra.get("error") is True

    check_results(sync_results)

    async def apredict(inputs: dict):
        return predict(inputs)

    with suppress_warnings():
        async_results = await aevaluate(
            apredict,
            data=list(
                client.list_examples(
                    dataset_name=dataset_name,
                    as_of="test_version",
                )
            ),
            evaluators=[
                error_dict_evaluator,
                error_evaluation_result,
                error_evaluation_results,
                error_dict_no_key,
                error_evaluation_results_dict,
            ],
            max_concurrency=1,  # To ensure deterministic order
        )

    assert len(async_results) == 10  # Assuming 10 examples in the dataset
    check_results([res async for res in async_results])


@functools.lru_cache(maxsize=1)
def _has_pandas() -> bool:
    try:
        import pandas  # noqa

        return True

    except Exception:
        return False


async def test_aevaluate():
    client = Client()
    dataset = client.clone_public_dataset(
        "https://smith.langchain.com/public/2bbf4a10-c3d5-4868-9e96-400df97fed69/d"
    )

    def accuracy(run: Run, example: Example):
        pred = run.outputs["output"]  # type: ignore
        expected = example.outputs["answer"]  # type: ignore
        return {"score": expected.lower() == pred.lower()}

    async def slow_accuracy(run: Run, example: Example):
        pred = run.outputs["output"]  # type: ignore
        expected = example.outputs["answer"]  # type: ignore
        await asyncio.sleep(2)
        return {"score": expected.lower() == pred.lower()}

    def precision(runs: Sequence[Run], examples: Sequence[Example]):
        predictions = [run.outputs["output"].lower() for run in runs]  # type: ignore
        expected = [example.outputs["answer"].lower() for example in examples]  # type: ignore
        tp = sum([p == e for p, e in zip(predictions, expected) if p == "yes"])
        fp = sum([p == "yes" and e == "no" for p, e in zip(predictions, expected)])
        return {"score": tp / (tp + fp)}

    async def apredict(inputs: dict) -> dict:
        await asyncio.sleep(0.1)
        return {"output": "Yes"}

    results = await aevaluate(
        apredict,
        data=dataset.name,
        evaluators=[accuracy, slow_accuracy],
        summary_evaluators=[precision],
        experiment_prefix="My Experiment",
        description="My Experiment Description",
        metadata={"my-prompt-version": "abcd-1234", "function": "aevaluate"},
    )
    assert len(results) == 10
    if _has_pandas():
        df = results.to_pandas()
        assert len(df) == 10
    async for _ in results:
        pass

    # Run it again with the existing project
    results2 = await aevaluate(
        apredict,
        data=dataset.name,
        evaluators=[accuracy],
        summary_evaluators=[precision],
        experiment=results.experiment_name,
    )
    assert len(results2) == 10

    # ... and again with the object
    experiment = client.read_project(project_name=results.experiment_name)
    results3 = await aevaluate(
        apredict,
        data=dataset.name,
        evaluators=[accuracy],
        summary_evaluators=[precision],
        experiment=experiment,
    )
    assert len(results3) == 10

    # ... and again with the ID
    results4 = await aevaluate(
        apredict,
        data=dataset.name,
        evaluators=[accuracy],
        summary_evaluators=[precision],
        experiment=str(experiment.id),
    )
    assert len(results4) == 10


def test_evaluate():
    client = Client()
    _ = client.clone_public_dataset(
        "https://smith.langchain.com/public/419dcab2-1d66-4b94-8901-0357ead390df/d"
    )
    dataset_name = "Evaluate Examples"

    def accuracy(run: Run, example: Example):
        pred = run.outputs["output"]  # type: ignore
        expected = example.outputs["answer"]  # type: ignore
        return {"score": expected.lower() == pred.lower()}

    def precision(runs: Sequence[Run], examples: Sequence[Example]):
        predictions = [run.outputs["output"].lower() for run in runs]  # type: ignore
        expected = [example.outputs["answer"].lower() for example in examples]  # type: ignore
        tp = sum([p == e for p, e in zip(predictions, expected) if p == "yes"])
        fp = sum([p == "yes" and e == "no" for p, e in zip(predictions, expected)])
        return {"score": tp / (tp + fp)}

    def predict(inputs: dict) -> dict:
        return {"output": "Yes"}

    results = evaluate(
        predict,
        data=client.list_examples(dataset_name=dataset_name, as_of="test_version"),
        evaluators=[accuracy],
        summary_evaluators=[precision],
        description="My sync experiment",
        metadata={
            "my-prompt-version": "abcd-1234",
            "function": "evaluate",
        },
        num_repetitions=3,
    )
    assert len(results) == 30
    if _has_pandas():
        df = results.to_pandas()
        assert len(df) == 30
        assert set(df.columns) == {
            "inputs.context",
            "inputs.question",
            "outputs.output",
            "error",
            "reference.answer",
            "feedback.accuracy",
            "execution_time",
            "example_id",
            "id",
        }
    examples = client.list_examples(dataset_name=dataset_name, as_of="test_version")
    for example in examples:
        assert len([r for r in results if r["example"].id == example.id]) == 3

    # Run it again with the existing project
    results2 = evaluate(
        predict,
        data=client.list_examples(dataset_name=dataset_name, as_of="test_version"),
        evaluators=[accuracy],
        summary_evaluators=[precision],
        experiment=results.experiment_name,
    )
    assert len(results2) == 10

    # ... and again with the object
    experiment = client.read_project(project_name=results.experiment_name)
    results3 = evaluate(
        predict,
        data=client.list_examples(dataset_name=dataset_name, as_of="test_version"),
        evaluators=[accuracy],
        summary_evaluators=[precision],
        experiment=experiment,
    )
    assert len(results3) == 10

    # ... and again with the ID
    results4 = evaluate(
        predict,
        data=client.list_examples(dataset_name=dataset_name, as_of="test_version"),
        evaluators=[accuracy],
        summary_evaluators=[precision],
        experiment=str(experiment.id),
    )
    assert len(results4) == 10


@test
def test_foo():
    expect(3 + 4).to_equal(7)


@pytest.fixture
def some_input():
    return "Some input"


@pytest.fixture
def expected_output():
    return "input"


@test(output_keys=["expected_output"])
def test_bar(some_input: str, expected_output: str):
    expect(some_input).to_contain(expected_output)


@test
async def test_baz():
    await asyncio.sleep(0.1)
    expect(3 + 4).to_equal(7)
    return 7


@test
@pytest.mark.parametrize("x, y", [(1, 2), (2, 3)])
def test_foo_parametrized(x, y):
    expect(x + y).to_be_greater_than(0)
    return x + y


@test(output_keys=["z"])
@pytest.mark.parametrize("x, y, z", [(1, 2, 3), (2, 3, 5)])
def test_bar_parametrized(x, y, z):
    expect(x + y).to_equal(z)
    return {"z": x + y}


@test(test_suite_name="tests.evaluation.test_evaluation::test_foo_async_parametrized")
@pytest.mark.parametrize("x, y", [(1, 2), (2, 3)])
async def test_foo_async_parametrized(x, y):
    await asyncio.sleep(0.1)
    expect(x + y).to_be_greater_than(0)
    return x + y


@test(output_keys=["z"])
@pytest.mark.parametrize("x, y, z", [(1, 2, 3), (2, 3, 5)])
async def test_bar_async_parametrized(x, y, z):
    await asyncio.sleep(0.1)
    expect(x + y).to_equal(z)
    return {"z": x + y}


@test
def test_pytest_skip():
    pytest.skip("Skip this test")


@test
async def test_async_pytest_skip():
    pytest.skip("Skip this test")


async def test_aevaluate_good_error():
    client = Client()
    ds_name = "__Empty Dataset Do Not Modify"
    if not client.has_dataset(dataset_name=ds_name):
        client.create_dataset(dataset_name=ds_name)

    async def predict(inputs: dict):
        return {}

    match_val = "No examples found in the dataset."
    with pytest.raises(ValueError, match=match_val):
        await aevaluate(
            predict,
            data=ds_name,
        )

    with pytest.raises(ValueError, match="Must specify 'data'"):
        await aevaluate(
            predict,
            data=[],
        )
    with pytest.raises(ValueError, match=match_val):
        await aevaluate(
            predict,
            data=(_ for _ in range(0)),
        )


async def test_aevaluate_large_dataset_and_concurrency():
    client = Client()
    _ = client.clone_public_dataset(
        "https://smith.langchain.com/public/2bbf4a10-c3d5-4868-9e96-400df97fed69/d"
    )
    dataset_name = "Evaluate Examples"

    async def mock_chat_completion(*, messages):
        await asyncio.sleep(1)
        return {
            "role": "assistant",
            "content": "Still thinking...",
        }

    def simulate_conversation_turn(*, existing, model_response):
        return existing + [
            model_response,
            {"role": "human", "content": "Think harder!"},
        ]

    # Will be traced by default
    async def target(inputs: dict) -> dict:
        messages = [
            {
                "role": "system",
                "content": "Come up with a math equation that solves the puzzle.",
            },
            # This dataset has inputs as a dict with a "statement" key
            {"role": "user", "content": "foo"},
        ]
        res = await mock_chat_completion(model="gpt-4o-mini", messages=messages)
        messages = simulate_conversation_turn(existing=messages, model_response=res)

        return {"equation": res}

    async def mock_evaluator_chat_completion(*, model, messages):
        await asyncio.sleep(2)
        return {
            "role": "assistant",
            "content": str(0.5),
        }

    async def mock_correctness_evaluator(outputs: dict, reference_outputs: dict):
        messages = [
            {"role": "system", "content": "Assign a score to the following output."},
            {
                "role": "user",
                "content": f"""
Actual: {outputs["equation"]}
""",
            },
        ]
        res = await mock_evaluator_chat_completion(model="o3-mini", messages=messages)
        return {
            "key": "correctness",
            "score": float(res["content"]),
            "comment": "The answer was a good attempt, but incorrect.",
        }

    client = Client()

    start = time.time()

    await client.aevaluate(
        target,
        data=client.list_examples(dataset_name=dataset_name, as_of="test_version"),
        evaluators=[
            mock_correctness_evaluator,
        ],
        max_concurrency=3,
    )

    finish_time = time.time()
    assert (finish_time - start) <= 8.5
